# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/registration_abci skill's behaviours."""

# pylint: skip-file

import collections
import datetime
import json
import logging
import time
from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, cast
from unittest import mock
from unittest.mock import MagicMock
from urllib.parse import urlparse

import pytest
from _pytest.logging import LogCaptureFixture

from packages.valory.contracts.service_registry.contract import ServiceRegistryContract
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.tendermint.message import TendermintMessage
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    TimeoutException,
    make_degenerate_behaviour,
)
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.registration_abci import PUBLIC_ID
from packages.valory.skills.registration_abci.behaviours import (
    RegistrationBaseBehaviour,
    RegistrationBehaviour,
    RegistrationStartupBehaviour,
)
from packages.valory.skills.registration_abci.models import SharedState
from packages.valory.skills.registration_abci.rounds import (
    BaseSynchronizedData as RegistrationSynchronizedData,
)
from packages.valory.skills.registration_abci.rounds import (
    Event,
    FinishedRegistrationRound,
)


PACKAGE_DIR = Path(__file__).parent.parent


SERVICE_REGISTRY_ADDRESS = "0xa51c1fc2f0d1a1b8494ed1fe312d7c3a78ed91c0"
ZERO_ADDRESS = "0x0000000000000000000000000000000000000000"
CONTRACT_ID = str(ServiceRegistryContract.contract_id)
ON_CHAIN_SERVICE_ID = 42
DUMMY_ADDRESS = "localhost"
DUMMY_VALIDATOR_CONFIG = {
    "hostname": DUMMY_ADDRESS,
    "address": "address",
    "pub_key": {
        "type": "tendermint/PubKeyEd25519",
        "value": "7y7ycBMMABj5Onf74ITYtUS3uZ6SsCQKZML87mIX",
    },
    "peer_id": "peer_id",
    "p2p_port": 80,
}


def test_skill_public_id() -> None:
    """Test skill module public ID"""

    assert PUBLIC_ID.name == Path(__file__).parents[1].name
    assert PUBLIC_ID.author == Path(__file__).parents[3].name


def consume(iterator: Iterable) -> None:
    """Consume the iterator"""
    collections.deque(iterator, maxlen=0)


@contextmanager
def as_context(*contexts: Any) -> Generator[None, None, None]:
    """Set contexts"""
    with ExitStack() as stack:
        consume(map(stack.enter_context, contexts))
        yield


class RegistrationAbciBaseCase(FSMBehaviourBaseCase):
    """Base case for testing RegistrationAbci FSMBehaviour."""

    path_to_skill = PACKAGE_DIR


class BaseRegistrationTestBehaviour(RegistrationAbciBaseCase):
    """Base test case to test RegistrationBehaviour."""

    behaviour_class = RegistrationBaseBehaviour
    next_behaviour_class = BaseBehaviour

    @pytest.mark.parametrize(
        "setup_data, expected_initialisation",
        (
            ({}, '{"0": {}}'),
            ({"test": []}, '{"0": {}}'),
            ({"test": [], "valid": [1, 2]}, '{"0": {"valid": [1, 2]}}'),
        ),
    )
    def test_registration(
        self, setup_data: Dict, expected_initialisation: Optional[str]
    ) -> None:
        """Test registration."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.auto_behaviour_id(),
            RegistrationSynchronizedData(AbciAppDB(setup_data=setup_data)),
        )
        assert isinstance(self.behaviour.current_behaviour, BaseBehaviour)
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )
        with mock.patch.object(
            self.behaviour.current_behaviour,
            "send_a2a_transaction",
            side_effect=self.behaviour.current_behaviour.send_a2a_transaction,
        ):
            self.behaviour.act_wrapper()
            assert isinstance(
                self.behaviour.current_behaviour.send_a2a_transaction, MagicMock
            )
            assert (
                self.behaviour.current_behaviour.send_a2a_transaction.call_args[0][
                    0
                ].initialisation
                == expected_initialisation
            )
            self.mock_a2a_transaction()

        self._test_done_flag_set()
        self.end_round(Event.DONE)
        assert (
            self.behaviour.current_behaviour.behaviour_id
            == self.next_behaviour_class.auto_behaviour_id()
        )


class TestRegistrationStartupBehaviour(RegistrationAbciBaseCase):
    """Test case to test RegistrationStartupBehaviour."""

    behaviour_class = RegistrationStartupBehaviour
    next_behaviour_class = make_degenerate_behaviour(FinishedRegistrationRound)

    other_agents: List[str] = ["0xAlice", "0xBob", "0xCharlie"]
    _time_in_future = datetime.datetime.now() + datetime.timedelta(hours=10)
    _time_in_past = datetime.datetime.now() - datetime.timedelta(hours=10)

    def setup(self, **kwargs: Any) -> None:
        """Setup"""
        super().setup(**kwargs)
        self.state.params.__dict__["sleep_time"] = 0.01
        self.state.params.__dict__["share_tm_config_on_startup"] = True

    def teardown(self, **kwargs: Any) -> None:
        """Teardown."""
        super().teardown(**kwargs)
        self.state.initial_tm_configs = {}

    @property
    def agent_instances(self) -> List[str]:
        """Agent instance addresses"""
        return [*self.other_agents, self.state.context.agent_address]

    @property
    def state(self) -> RegistrationStartupBehaviour:
        """Current behavioural state"""
        return cast(RegistrationStartupBehaviour, self.behaviour.current_behaviour)

    @property
    def logger(self) -> str:
        """Logger"""
        return "aea.test_agent_name.packages.valory.skills.registration_abci"

    # mock patches
    @property
    def mocked_service_registry_address(self) -> mock._patch_dict:
        """Mocked service registry address"""
        return mock.patch.dict(
            self.state.params.__dict__,
            {"service_registry_address": SERVICE_REGISTRY_ADDRESS},
        )

    @property
    def mocked_on_chain_service_id(self) -> mock._patch_dict:
        """Mocked on chain service id"""
        return mock.patch.dict(
            self.state.params.__dict__, {"on_chain_service_id": ON_CHAIN_SERVICE_ID}
        )

    def mocked_wait_for_condition(self, should_timeout: bool) -> mock._patch:
        """Mock BaseBehaviour.wait_for_condition"""

        def dummy_wait_for_condition(
            condition: Callable[[], bool], timeout: Optional[float] = None
        ) -> Generator[None, None, None]:
            """A mock implementation of BaseBehaviour.wait_for_condition"""
            # call the condition
            condition()
            if should_timeout:
                # raise in case required
                raise TimeoutException()
            return
            yield

        return mock.patch.object(
            self.behaviour.current_behaviour,
            "wait_for_condition",
            side_effect=dummy_wait_for_condition,
        )

    @property
    def mocked_yield_from_sleep(self) -> mock._patch:
        """Mock yield from sleep"""
        return mock.patch.object(self.behaviour.current_behaviour, "sleep")

    # mock contract calls
    def mock_is_correct_contract(self, error_response: bool = False) -> None:
        """Mock service registry contract call to for contract verification"""
        request_kwargs = dict(performative=ContractApiMessage.Performative.GET_STATE)
        state = ContractApiMessage.State(ledger_id="ethereum", body={"verified": True})
        performative = ContractApiMessage.Performative.STATE
        if error_response:
            performative = ContractApiMessage.Performative.ERROR
        response_kwargs = dict(
            performative=performative,
            callable="verify_contract",
            state=state,
        )
        self.mock_contract_api_request(
            contract_id=CONTRACT_ID,
            request_kwargs=request_kwargs,
            response_kwargs=response_kwargs,
        )

    def mock_get_agent_instances(
        self, *agent_instances: str, error_response: bool = False
    ) -> None:
        """Mock get agent instances"""
        request_kwargs = dict(performative=ContractApiMessage.Performative.GET_STATE)
        performative = ContractApiMessage.Performative.STATE
        if error_response:
            performative = ContractApiMessage.Performative.ERROR
        body = {"agentInstances": list(agent_instances)}
        state = ContractApiMessage.State(ledger_id="ethereum", body=body)
        response_kwargs = dict(
            performative=performative,
            callable="get_agent_instances",
            state=state,
        )
        self.mock_contract_api_request(
            contract_id=CONTRACT_ID,
            request_kwargs=request_kwargs,
            response_kwargs=response_kwargs,
        )

    # mock Tendermint config request
    def mock_tendermint_request(
        self, request_kwargs: Dict, response_kwargs: Dict
    ) -> None:
        """Mock Tendermint request."""

        actual_tendermint_message = self.get_message_from_outbox()
        assert actual_tendermint_message is not None, "No message in outbox."
        has_attributes, error_str = self.message_has_attributes(
            actual_message=actual_tendermint_message,
            message_type=TendermintMessage,
            performative=TendermintMessage.Performative.GET_GENESIS_INFO,
            sender=self.state.context.agent_address,
            to=actual_tendermint_message.to,
            **request_kwargs,
        )
        assert has_attributes, error_str
        incoming_message = self.build_incoming_message(
            message_type=TendermintMessage,
            dialogue_reference=(
                actual_tendermint_message.dialogue_reference[0],
                "stub",
            ),
            performative=TendermintMessage.Performative.GENESIS_INFO,
            target=actual_tendermint_message.message_id,
            message_id=-1,
            to=self.state.context.agent_address,
            sender=actual_tendermint_message.to,
            **response_kwargs,
        )
        self.tendermint_handler.handle(cast(TendermintMessage, incoming_message))

    def mock_get_tendermint_info(self, *addresses: str) -> None:
        """Mock get Tendermint info"""
        for _ in addresses:
            request_kwargs: Dict = dict()
            info = json.dumps(DUMMY_VALIDATOR_CONFIG)
            response_kwargs = dict(info=info)
            self.mock_tendermint_request(request_kwargs, response_kwargs)
        # give room to the behaviour to finish sleeping
        # using the same sleep time here and in the behaviour
        # can lead to problems when the sleep here is finished
        # before the one in the behaviour
        time.sleep(self.state.params.sleep_time * 2)
        self.behaviour.act_wrapper()

    # mock HTTP requests
    def mock_get_local_tendermint_params(self, valid_response: bool = True) -> None:
        """Mock Tendermint get local params"""
        url = self.state.tendermint_parameter_url
        request_kwargs = dict(method="GET", url=url)
        body = b""
        if valid_response:
            params = dict(params=DUMMY_VALIDATOR_CONFIG, status=True, error=None)
            body = json.dumps(params).encode(self.state.ENCODING)
        response_kwargs = dict(status_code=200, body=body)
        self.mock_http_request(request_kwargs, response_kwargs)

    def mock_tendermint_update(self, valid_response: bool = True) -> None:
        """Mock Tendermint update"""

        validator_configs = self.state.format_genesis_data(
            self.state.initial_tm_configs
        )
        body = json.dumps(validator_configs).encode(self.state.ENCODING)
        url = self.state.tendermint_parameter_url
        request_kwargs = dict(method="POST", url=url, body=body)
        body = (
            json.dumps({"status": True, "error": None}).encode(self.state.ENCODING)
            if valid_response
            else b""
        )
        response_kwargs = dict(status_code=200, body=body)
        self.mock_http_request(request_kwargs, response_kwargs)

    def set_last_timestamp(self, last_timestamp: Optional[datetime.datetime]) -> None:
        """Set last timestamp"""
        if last_timestamp is not None:
            state = cast(SharedState, self._skill.skill_context.state)
            state.round_sequence.blockchain._blocks.append(
                MagicMock(timestamp=last_timestamp)
            )

    @staticmethod
    def dummy_reset_tendermint_with_wait_wrapper(
        valid_response: bool,
    ) -> Callable[[], Generator[None, None, Optional[bool]]]:
        """Wrapper for a Dummy `reset_tendermint_with_wait` method."""

        def dummy_reset_tendermint_with_wait(
            **_: bool,
        ) -> Generator[None, None, Optional[bool]]:
            """Dummy `reset_tendermint_with_wait` method."""
            yield
            return valid_response

        return dummy_reset_tendermint_with_wait

    # tests
    def test_init(self) -> None:
        """Empty init"""
        assert self.state.initial_tm_configs == {}
        assert self.state.local_tendermint_params == {}
        assert self.state.updated_genesis_data == {}

    def test_no_contract_address(self, caplog: LogCaptureFixture) -> None:
        """Test service registry contract address not provided"""
        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_yield_from_sleep,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            log_message = self.state.LogMessages.no_contract_address
        assert log_message.value in caplog.text

    @pytest.mark.parametrize("valid_response", [True, False])
    def test_request_personal(
        self, valid_response: bool, caplog: LogCaptureFixture
    ) -> None:
        """Test get tendermint configuration"""

        failed_message = self.state.LogMessages.failed_personal
        response_message = self.state.LogMessages.response_personal
        log_message = [failed_message, response_message][valid_response]
        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_yield_from_sleep,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params(valid_response=valid_response)
            assert log_message.value in caplog.text

    def test_failed_verification(self, caplog: LogCaptureFixture) -> None:
        """Test service registry contract not correctly deployed"""

        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract(error_response=True)
            log_message = self.state.LogMessages.failed_verification
            assert log_message.value in caplog.text

    def test_on_chain_service_id_not_set(self, caplog: LogCaptureFixture) -> None:
        """Test `get_addresses` when `on_chain_service_id` is `None`."""

        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            log_message = self.state.LogMessages.no_on_chain_service_id
            assert log_message.value in caplog.text

    def test_failed_service_info(self, caplog: LogCaptureFixture) -> None:
        """Test get service info failure"""

        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_on_chain_service_id,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            self.mock_get_agent_instances(error_response=True)
            log_message = self.state.LogMessages.failed_service_info
            assert log_message.value in caplog.text

    def test_no_agents_registered(self, caplog: LogCaptureFixture) -> None:
        """Test no agent instances registered"""

        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_on_chain_service_id,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            self.mock_get_agent_instances()
            log_message = self.state.LogMessages.no_agents_registered
            assert log_message.value in caplog.text

    def test_self_not_registered(self, caplog: LogCaptureFixture) -> None:
        """Test node operator agent not registered"""

        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_on_chain_service_id,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            self.mock_get_agent_instances(*self.other_agents)
            log_message = self.state.LogMessages.self_not_registered
            assert log_message.value in caplog.text

    def test_response_service_info(self, caplog: LogCaptureFixture) -> None:
        """Test registered addresses retrieved"""

        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_on_chain_service_id,
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            self.mock_get_agent_instances(*self.agent_instances)

            assert set(self.state.initial_tm_configs) == set(self.agent_instances)
            my_info = self.state.initial_tm_configs[self.state.context.agent_address]
            assert (
                my_info["hostname"]
                == urlparse(self.state.context.params.tendermint_url).hostname
            )
            assert not any(map(self.state.initial_tm_configs.get, self.other_agents))
            log_message = self.state.LogMessages.response_service_info
            assert log_message.value in caplog.text

    def test_collection_complete(self, caplog: LogCaptureFixture) -> None:
        """Test registered addresses retrieved"""

        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_on_chain_service_id,
            mock.patch.object(
                self._skill.skill_context.state,
                "acn_container",
                side_effect=lambda: self.agent_instances,
            ),
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            self.mock_get_agent_instances(*self.agent_instances)
            self.mock_get_tendermint_info(*self.other_agents)
            assert all(map(self.state.initial_tm_configs.get, self.other_agents))
            log_message = self.state.LogMessages.collection_complete
            assert log_message.value in caplog.text

    @pytest.mark.parametrize("valid_response", [True, False])
    @mock.patch.object(BaseBehaviour, "reset_tendermint_with_wait")
    def test_request_update(
        self, _: mock.Mock, valid_response: bool, caplog: LogCaptureFixture
    ) -> None:
        """Test Tendermint config update"""

        self.state.updated_genesis_data = {}
        failed_message = self.state.LogMessages.failed_update
        response_message = self.state.LogMessages.response_update
        log_message = [failed_message, response_message][valid_response]
        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_on_chain_service_id,
            self.mocked_yield_from_sleep,
            mock.patch.object(
                self._skill.skill_context.state,
                "acn_container",
                side_effect=lambda: self.agent_instances,
            ),
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            self.mock_get_agent_instances(*self.agent_instances)
            self.mock_get_tendermint_info(*self.other_agents)
            self.mock_tendermint_update(valid_response)
            assert log_message.value in caplog.text

    @pytest.mark.parametrize(
        "valid_response, last_timestamp, timeout",
        [
            (True, _time_in_past, False),
            (False, _time_in_past, False),
            (True, _time_in_future, False),
            (False, _time_in_future, False),
            (True, None, False),
            (False, None, False),
            (True, None, True),
            (False, None, True),
        ],
    )
    def test_request_restart(
        self,
        valid_response: bool,
        last_timestamp: Optional[datetime.datetime],
        timeout: bool,
        caplog: LogCaptureFixture,
    ) -> None:
        """Test Tendermint start"""
        self.state.updated_genesis_data = {}
        self.set_last_timestamp(last_timestamp)
        with as_context(
            caplog.at_level(logging.INFO, logger=self.logger),
            self.mocked_service_registry_address,
            self.mocked_on_chain_service_id,
            self.mocked_wait_for_condition(timeout),
            self.mocked_yield_from_sleep,
            mock.patch.object(
                self.behaviour.current_behaviour,
                "reset_tendermint_with_wait",
                side_effect=self.dummy_reset_tendermint_with_wait_wrapper(
                    valid_response
                ),
            ),
            mock.patch.object(
                self._skill.skill_context.state,
                "acn_container",
                side_effect=lambda: self.agent_instances,
            ),
        ):
            self.behaviour.act_wrapper()
            self.mock_get_local_tendermint_params()
            self.mock_is_correct_contract()
            self.mock_get_agent_instances(*self.agent_instances)
            self.mock_get_tendermint_info(*self.other_agents)
            self.mock_tendermint_update()
            self.behaviour.act_wrapper()


class TestRegistrationStartupBehaviourNoConfigShare(BaseRegistrationTestBehaviour):
    """Test case to test RegistrationBehaviour."""

    behaviour_class = RegistrationStartupBehaviour
    next_behaviour_class = make_degenerate_behaviour(FinishedRegistrationRound)


class TestRegistrationBehaviour(BaseRegistrationTestBehaviour):
    """Test case to test RegistrationBehaviour."""

    behaviour_class = RegistrationBehaviour
    next_behaviour_class = make_degenerate_behaviour(FinishedRegistrationRound)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test base."""

import os
import shutil
import subprocess  # nosec
import sys
import tempfile
from contextlib import suppress
from pathlib import Path
from typing import Optional, Sequence, Tuple

import pytest
from _pytest.capture import CaptureFixture  # type: ignore
from aea.test_tools.click_testing import CliRunner
from click.testing import Result

from autonomy.cli import cli


class BaseCliTest:
    """Test `autonomy analyse abci` command."""

    @pytest.fixture(autouse=True)
    def set_capfd_on_cli_runner(self, capfd: CaptureFixture) -> None:
        """Set pytest capfd on CLI runner"""
        self.cli_runner.capfd = capfd

    t: Path
    cwd: Path
    cli_runner: CliRunner
    cli_options: Tuple[str, ...]

    @classmethod
    def setup_class(cls) -> None:
        """Setup test class."""
        cls.cli_runner = CliRunner()
        cls.cwd = Path.cwd().absolute()

    def setup(
        self,
    ) -> None:
        """Setup test."""
        self.t = Path(tempfile.mkdtemp())

    def run_cli(self, commands: Optional[Tuple[str, ...]] = None) -> Result:
        """Run CLI."""
        if commands is None:
            return self.cli_runner.invoke(cli=cli, args=self.cli_options)

        return self.cli_runner.invoke(cli=cli, args=(*self.cli_options, *commands))

    def run_cli_subprocess(
        self, commands: Sequence[str], timeout: float = 60.0
    ) -> Tuple[int, str, str]:
        """Run CLI using subprocess."""
        process = subprocess.Popen(  # nosec
            [sys.executable, "-m", "autonomy.cli"]
            + list(self.cli_options)
            + list(commands),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout_bytes, stderr_bytes = process.communicate(timeout=timeout)
        stdout, stderr = stdout_bytes.decode(), stderr_bytes.decode()

        # for Windows
        if sys.platform == "win32":
            stdout = stdout.replace("\r", "")
            stderr = stderr.replace("\r", "")
        return process.returncode, stdout, stderr

    def teardown(
        self,
    ) -> None:
        """Teardown method."""
        os.chdir(self.cwd)
        with suppress(OSError, FileExistsError, PermissionError):
            shutil.rmtree(str(self.t))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for the `IPFS` interactions."""

# pylint: skip-file

from typing import Dict, cast
from unittest import mock

import pytest

from packages.valory.skills.abstract_round_abci.io_.ipfs import (
    IPFSInteract,
    IPFSInteractionError,
)
from packages.valory.skills.abstract_round_abci.io_.load import AbstractLoader
from packages.valory.skills.abstract_round_abci.io_.store import (
    AbstractStorer,
    StoredJSONType,
    SupportedFiletype,
)


use_ipfs_daemon = pytest.mark.usefixtures("ipfs_daemon")


class TestIPFSInteract:
    """Test `IPFSInteract`."""

    def setup(self) -> None:
        """Setup test class."""
        self.ipfs_interact = IPFSInteract()

    @pytest.mark.parametrize("multiple", (True, False))
    def test_store_and_send_and_back(
        self,
        multiple: bool,
        dummy_obj: StoredJSONType,
        dummy_multiple_obj: Dict[str, StoredJSONType],
    ) -> None:
        """Test store -> send -> download -> read of objects."""
        obj: StoredJSONType
        if multiple:
            obj = dummy_multiple_obj
            filepath = "dummy_dir"
        else:
            obj = dummy_obj
            filepath = "test_file.json"

        serialized_objects = self.ipfs_interact.store(
            filepath, obj, multiple, SupportedFiletype.JSON
        )
        expected_objects = obj
        actual_objects = cast(
            Dict[str, str],
            self.ipfs_interact.load(
                serialized_objects,
                SupportedFiletype.JSON,
            ),
        )
        if multiple:
            # here we manually remove the trailing the dir from the name.
            # This is done by the IPFS connection under normal circumstances.
            actual_objects = {
                k.lstrip(filepath).lstrip("/").lstrip("\\"): v
                for k, v in actual_objects.items()
            }

        assert actual_objects == expected_objects

    def test_store_fails(self, dummy_multiple_obj: Dict[str, StoredJSONType]) -> None:
        """Tests when "store" fails."""
        dummy_filepath = "dummy_dir"
        multiple = False
        with mock.patch.object(
            AbstractStorer,
            "store",
            side_effect=ValueError,
        ), pytest.raises(IPFSInteractionError):
            self.ipfs_interact.store(
                dummy_filepath, dummy_multiple_obj, multiple, SupportedFiletype.JSON
            )

    def test_load_fails(self, dummy_multiple_obj: Dict[str, StoredJSONType]) -> None:
        """Tests when "load" fails."""
        dummy_object = {"test": "test"}
        with mock.patch.object(
            AbstractLoader,
            "load",
            side_effect=ValueError,
        ), pytest.raises(IPFSInteractionError):
            self.ipfs_interact.load(dummy_object)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'abci' skill."""

from typing import Generator, Set, Type, cast

from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.register_reset_recovery_abci.composition import (
    RegisterResetRecoveryAbciApp,
)
from packages.valory.skills.register_reset_recovery_abci.models import SharedState
from packages.valory.skills.register_reset_recovery_abci.payloads import (
    RoundCountPayload,
)
from packages.valory.skills.register_reset_recovery_abci.rounds import RoundCountRound
from packages.valory.skills.registration_abci.behaviours import (
    AgentRegistrationRoundBehaviour,
    RegistrationStartupBehaviour,
)


class ShareRoundCountBehaviour(BaseBehaviour):
    """A dummy behaviour that sync the round count."""

    matching_round = RoundCountRound

    def async_act(self) -> Generator:
        """Simply logs the round count, and share it."""
        round_count: int = cast(
            SharedState, self.context.state
        ).synchronized_data.round_count

        self.context.logger.info(f"Current round count is {round_count}. ")
        payload = RoundCountPayload(
            sender=self.context.agent_address, current_round_count=round_count
        )
        # the following sleep acts as a buffer so that
        # we are sure that all agents have enough time to print the log above
        # in contrary, it might happen that majority is reached before 1 agent
        # has printed the round count log, which might lead to a failing test.
        # This is because a new round is scheduled as soon as the desired majority
        # is reached. There is no guarantee that all behaviours across all agents
        # will finish given an `CollectSameUntilThresholdRound` round.
        buffer = 5
        yield from self.sleep(buffer)
        yield from self.send_a2a_transaction(payload)
        yield from self.wait_until_round_end()
        self.set_done()


class RegisterResetRecoveryAbciAppConsensusBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the register-reset-recovery."""

    initial_behaviour_cls = RegistrationStartupBehaviour
    abci_app_cls = RegisterResetRecoveryAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        *AgentRegistrationRoundBehaviour.behaviours,
        ShareRoundCountBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the rounds of the skill."""

# pylint: skip-file

import logging  # noqa: F401
from typing import cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciAppDB,
    CollectionRound,
    MAX_INT_256,
)
from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
    BaseRoundTestClass as ExternalBaseRoundTestClass,
)
from packages.valory.skills.hello_world_abci.payloads import (
    CollectRandomnessPayload,
    PrintMessagePayload,
    RegistrationPayload,
    ResetPayload,
    SelectKeeperPayload,
)
from packages.valory.skills.hello_world_abci.rounds import (
    CollectRandomnessRound,
    Event,
    PrintMessageRound,
    RegistrationRound,
    ResetAndPauseRound,
    SelectKeeperRound,
    SynchronizedData,
)


MAX_PARTICIPANTS: int = 4
RANDOMNESS: str = "d1c29dce46f979f9748210d24bce4eae8be91272f5ca1a6aea2832d3dd676f51"


class BaseRoundTestClass(ExternalBaseRoundTestClass):
    """Base test class for Rounds."""

    _synchronized_data_class = SynchronizedData
    _event_class = Event


class TestRegistrationRound(BaseRoundTestClass):
    """Tests for RegistrationRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = RegistrationRound(
            synchronized_data=self.synchronized_data,
        )

        first_payload, *payloads = [
            RegistrationPayload(sender=participant) for participant in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection == {first_payload.sender: first_payload}
        assert test_round.end_block() is None

        for payload in payloads:
            test_round.process_payload(payload)

        actual_next_behaviour = SynchronizedData(
            AbciAppDB(setup_data=dict(participants=[tuple(test_round.collection)]))
        )

        res = test_round.end_block()
        assert res is not None
        synchronized_data, event = res
        assert (
            cast(SynchronizedData, synchronized_data).participants
            == cast(SynchronizedData, actual_next_behaviour).participants
        )
        assert event == Event.DONE


class TestCollectRandomnessRound(BaseRoundTestClass):
    """Tests for CollectRandomnessRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = CollectRandomnessRound(
            synchronized_data=self.synchronized_data,
        )
        first_payload, *payloads = [
            CollectRandomnessPayload(
                sender=participant, randomness=RANDOMNESS, round_id=0
            )
            for participant in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection[first_payload.sender] == first_payload
        assert test_round.end_block() is None

        self._test_no_majority_event(test_round)

        for payload in payloads:
            test_round.process_payload(payload)

        actual_next_behaviour = self.synchronized_data.update(
            participant_to_randomness=test_round.serialized_collection,
            most_voted_randomness=test_round.most_voted_payload,
        )

        res = test_round.end_block()
        assert res is not None
        synchronized_data, event = res
        assert all(
            [
                key
                in cast(SynchronizedData, synchronized_data).participant_to_randomness
                for key in cast(
                    SynchronizedData, actual_next_behaviour
                ).participant_to_randomness
            ]
        )
        assert event == Event.DONE


class TestSelectKeeperRound(BaseRoundTestClass):
    """Tests for SelectKeeperRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = SelectKeeperRound(
            synchronized_data=self.synchronized_data,
        )

        first_payload, *payloads = [
            SelectKeeperPayload(sender=participant, keeper="keeper")
            for participant in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection[first_payload.sender] == first_payload
        assert test_round.end_block() is None

        self._test_no_majority_event(test_round)

        for payload in payloads:
            test_round.process_payload(payload)

        actual_next_behaviour = self.synchronized_data.update(
            participant_to_selection=test_round.serialized_collection,
            most_voted_keeper_address=test_round.most_voted_payload,
        )

        res = test_round.end_block()
        assert res is not None
        synchronized_data, event = res
        assert all(
            [
                key
                in cast(SynchronizedData, synchronized_data).participant_to_selection
                for key in cast(
                    SynchronizedData, actual_next_behaviour
                ).participant_to_selection
            ]
        )
        assert event == Event.DONE


class TestPrintMessageRound(BaseRoundTestClass):
    """Tests for PrintMessageRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = PrintMessageRound(
            synchronized_data=self.synchronized_data,
        )

        first_payload, *payloads = [
            PrintMessagePayload(sender=participant, message=f"{participant}_message")
            for participant in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection == {first_payload.sender: first_payload}
        assert test_round.end_block() is None

        for payload in payloads:
            test_round.process_payload(payload)

        printed_messages = [
            cast(PrintMessagePayload, payload).message
            for payload in test_round.collection.values()
        ]

        actual_next_behaviour = SynchronizedData(
            AbciAppDB(
                setup_data=dict(
                    participants=[tuple(sorted(test_round.collection))],
                    printed_messages=[sorted(printed_messages)],
                )
            )
        )

        res = test_round.end_block()
        assert res is not None
        synchronized_data, event = res

        assert (
            cast(SynchronizedData, synchronized_data).participants
            == cast(SynchronizedData, actual_next_behaviour).participants
        )
        assert (
            cast(SynchronizedData, synchronized_data).printed_messages
            == cast(SynchronizedData, actual_next_behaviour).printed_messages
        )
        assert event == Event.DONE


class TestResetAndPauseRound(BaseRoundTestClass):
    """Tests for ResetAndPauseRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""
        test_round = ResetAndPauseRound(
            synchronized_data=self.synchronized_data,
        )

        first_payload, *payloads = [
            ResetPayload(sender=participant, period_count=1)
            for participant in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection[first_payload.sender] == first_payload
        assert test_round.end_block() is None

        self._test_no_majority_event(test_round)

        for payload in payloads:
            test_round.process_payload(payload)

        actual_next_behaviour = self.synchronized_data.create()

        res = test_round.end_block()
        assert res is not None
        synchronized_data, event = res

        assert (
            cast(SynchronizedData, synchronized_data).period_count
            == cast(SynchronizedData, actual_next_behaviour).period_count
        )

        assert event == Event.DONE


def test_synchronized_data() -> None:  # pylint:too-many-locals
    """Test SynchronizedData."""

    participants = tuple(f"agent_{i}" for i in range(MAX_PARTICIPANTS))
    participant_to_randomness = {
        participant: CollectRandomnessPayload(
            sender=participant, randomness=RANDOMNESS, round_id=0
        )
        for participant in participants
    }
    participant_to_randomness_serialized = CollectionRound.serialize_collection(
        participant_to_randomness
    )
    most_voted_randomness = "0xabcd"
    participant_to_selection = {
        participant: SelectKeeperPayload(sender=participant, keeper="keeper")
        for participant in participants
    }
    participant_to_selection_serialized = CollectionRound.serialize_collection(
        participant_to_selection
    )
    most_voted_keeper_address = "keeper"

    synchronized_data = SynchronizedData(
        AbciAppDB(
            setup_data=AbciAppDB.data_to_lists(
                dict(
                    participants=participants,
                    setup_params={},
                    participant_to_randomness=participant_to_randomness_serialized,
                    most_voted_randomness=most_voted_randomness,
                    participant_to_selection=participant_to_selection_serialized,
                    most_voted_keeper_address=most_voted_keeper_address,
                )
            ),
        )
    )

    assert synchronized_data.participants == frozenset(participants)
    assert synchronized_data.period_count == 0
    assert synchronized_data.participant_to_randomness == participant_to_randomness
    assert synchronized_data.most_voted_randomness == most_voted_randomness
    assert synchronized_data.participant_to_selection == participant_to_selection
    assert synchronized_data.most_voted_keeper_address == most_voted_keeper_address
    assert synchronized_data.sorted_participants == sorted(participants)
    actual_keeper_randomness = int(most_voted_randomness, base=16) / MAX_INT_256
    assert (
        abs(synchronized_data.keeper_randomness - actual_keeper_randomness) < 1e-10
    )  # avoid equality comparisons between floats



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the class to connect to the Service Registry contract."""

import hashlib
import json
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

from aea.common import JSONLike
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi


PUBLIC_ID = PublicId.from_str("valory/service_registry:0.1.0")
ETHEREUM_IDENTIFIER = "ethereum"
UNIT_HASH_PREFIX = "0x{metadata_hash}"
EXPECTED_CONTRACT_ADDRESS_BY_CHAIN_ID = {
    1: "0x48b6af7B12C71f09e2fC8aF4855De4Ff54e775cA",
    5: "0x1cEe30D08943EB58EFF84DD1AB44a6ee6FEff63a",
    100: "0x9338b5153AE39BB89f50468E608eD9d764B755fD",
    137: "0xE3607b00E75f6405248323A9417ff6b39B244b50",
    31337: "0x998abeb3E57409262aE5b751f60747921B33613E",
}
DEPLOYED_BYTECODE_MD5_HASH_BY_CHAIN_ID = {
    1: "6f9fc7f3c2348801737120e6b5f8fa8e9670c65152c66d128ff4cddb465b4d705340c559e352f5e7f29bda3b84a8d36d4a9448b791cfe2d370e31c01276e0244",
    5: "d4a860f21f17762c99d93359244b39a878dd5bac9ea6056c0ff29c7558d6653aa0d5962aa819fc9f05f237d068845125cfc37a7fd7761b11c29a709ad5c48157",
    100: "10e2cfb500481d6c5a3b6b90507e4ac04d8b0d88741cea5568306ed4115f24e8b9747055423da5fca05838d5ccefebf41fb47d2ba1fb45215b6b21c0a27823be",
    137: "10e2cfb500481d6c5a3b6b90507e4ac04d8b0d88741cea5568306ed4115f24e8b9747055423da5fca05838d5ccefebf41fb47d2ba1fb45215b6b21c0a27823be",
    31337: "41ab54d43fd4bdfdc929658a0dc9bedd970c7339eecafbefda9892ab54c02c396bceedaa3a84a0f4690bee03dc11195a3267a264de7859c420efbf4291f1fef0",
}
L1_CHAINS = (
    1,
    5,
    31337,
)
L2_BUILD_FILENAME = "ServiceRegistryL2.json"

ServiceInfo = Tuple[int, str, bytes, int, int, int, int, List[int]]

_logger = logging.getLogger(
    f"aea.packages.{PUBLIC_ID.author}.contracts.{PUBLIC_ID.name}.contract"
)


class ServiceRegistryContract(Contract):
    """The Service Registry contract."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Get the Safe transaction."""
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[bytes]:
        """Get raw message."""
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> Optional[JSONLike]:
        """Get state."""
        raise NotImplementedError  # pragma: nocover

    @staticmethod
    def is_l1_chain(ledger_api: LedgerApi) -> bool:
        """Check if we're interecting with an L1 chain"""
        return ledger_api.api.eth.chain_id in L1_CHAINS

    @staticmethod
    def load_l2_build() -> JSONLike:
        """Load L2 ABI"""
        path = Path(__file__).parent / "build" / L2_BUILD_FILENAME
        return json.loads(path.read_text(encoding="utf-8"))

    @classmethod
    def get_instance(
        cls,
        ledger_api: LedgerApi,
        contract_address: Optional[str] = None,
    ) -> Any:
        """Get contract instance."""
        if ledger_api.identifier != ETHEREUM_IDENTIFIER:
            return super().get_instance(
                ledger_api=ledger_api, contract_address=contract_address
            )
        if cls.is_l1_chain(ledger_api=ledger_api):
            contract_interface = cls.contract_interface.get(ledger_api.identifier, {})
        else:
            contract_interface = cls.load_l2_build()
        instance = ledger_api.get_contract_instance(
            contract_interface, contract_address
        )
        return instance

    @classmethod
    def verify_contract(
        cls, ledger_api: LedgerApi, contract_address: str
    ) -> Dict[str, Union[bool, str]]:
        """
        Verify the contract's bytecode

        :param ledger_api: the ledger API object
        :param contract_address: the contract address
        :return: the verified status
        """
        verified = False
        chain_id = ledger_api.api.eth.chain_id
        expected_address = EXPECTED_CONTRACT_ADDRESS_BY_CHAIN_ID[chain_id]
        if contract_address != expected_address:
            _logger.error(
                f"For chain_id {chain_id} expected {expected_address} and got {contract_address}."
            )
            return dict(verified=verified)
        deployed_bytecode = ledger_api.api.eth.get_code(contract_address).hex()
        sha512_hash = hashlib.sha512(deployed_bytecode.encode("utf-8")).hexdigest()
        verified = DEPLOYED_BYTECODE_MD5_HASH_BY_CHAIN_ID[chain_id] == sha512_hash
        if not verified:
            _logger.error(
                f"CONTRACT NOT VERIFIED! Contract address: {contract_address}, chain_id: {chain_id}."
            )
        return dict(verified=verified)

    @classmethod
    def exists(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> bool:
        """Check if the service id exists"""

        contract_instance = cls.get_instance(ledger_api, contract_address)
        exists = ledger_api.contract_method_call(
            contract_instance=contract_instance,
            method_name="exists",
            unitId=service_id,
        )

        return cast(bool, exists)

    @classmethod
    def get_agent_instances(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> Dict[str, Any]:
        """Retrieve on-chain agent instances."""

        contract_instance = cls.get_instance(ledger_api, contract_address)
        service_info = ledger_api.contract_method_call(
            contract_instance=contract_instance,
            method_name="getAgentInstances",
            serviceId=service_id,
        )

        return dict(
            numAgentInstances=service_info[0],
            agentInstances=service_info[1],
        )

    @classmethod
    def get_service_owner(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> Dict[str, Any]:
        """Retrieve the service owner."""
        contract_instance = cls.get_instance(ledger_api, contract_address)
        service_owner = contract_instance.functions.ownerOf(service_id).call()
        checksum_service_owner = ledger_api.api.toChecksumAddress(service_owner)
        return dict(
            service_owner=checksum_service_owner,
        )

    @classmethod
    def get_service_information(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        token_id: int,
    ) -> ServiceInfo:
        """Retrieve service information"""

        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        return contract_interface.functions.getService(token_id).call()

    @classmethod
    def get_token_uri(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        token_id: int,
    ) -> str:
        """Resolve token URI"""

        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )
        return contract_interface.functions.tokenURI(token_id).call()

    @classmethod
    def filter_token_id_from_emitted_events(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
    ) -> Optional[int]:
        """Returns `CreateUnit` event filter."""

        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )

        events = contract_interface.events.CreateService.createFilter(
            fromBlock="latest"
        ).get_all_entries()

        for event in events:
            event_args = event["args"]
            if "serviceId" in event_args:
                return cast(int, event_args["serviceId"])

        return None

    @classmethod
    def verify_service_has_been_activated(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> bool:
        """Checks for a successful service registration event in the latest block"""

        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )

        events = contract_interface.events.ActivateRegistration.createFilter(
            fromBlock="latest"
        ).get_all_entries()
        for event in events:
            if event["args"]["serviceId"] == service_id:
                return True

        return False

    @classmethod
    def verify_agent_instance_registration(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
        instance_check: Set[str],
    ) -> Set[str]:
        """Checks for the registered instances and filters out the instances that are registered from the given array"""

        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )

        events = contract_interface.events.RegisterInstance.createFilter(
            fromBlock="latest"
        ).get_all_entries()

        successful = set()
        for event in events:
            event_args = event["args"]
            if event_args["serviceId"] != service_id:
                continue

            agent_instance = event_args["agentInstance"]
            if agent_instance in instance_check:
                successful.add(agent_instance)

        return successful

    @classmethod
    def verify_service_has_been_deployed(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        service_id: int,
    ) -> bool:
        """Checks for a successful service registration event in the latest block"""

        contract_interface = cls.get_instance(
            ledger_api=ledger_api,
            contract_address=contract_address,
        )

        events = contract_interface.events.DeployService.createFilter(
            fromBlock="latest"
        ).get_all_entries()
        for event in events:
            if event["args"]["serviceId"] == service_id:
                return True

        return False



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test hash command group."""


import os
import shutil
from pathlib import Path
from typing import Tuple
from unittest import mock

from autonomy.cli.helpers.registry import IPFSTool, REMOTE_IPFS, to_v1

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


class TestPublish(BaseCliTest):
    """Test `publish` command."""

    packages_dir: Path
    agent_dir: Path
    service_dir: Path
    temp_local_registry: Path

    cli_options: Tuple[str, ...] = ("publish",)

    def setup(self) -> None:
        """Setup class."""

        super().setup()

        self.packages_dir = self.t / "packages"
        self.service_dir = self.t / "packages" / "valory" / "services" / "counter"
        self.temp_local_registry = self.t / "temp_reg"
        self.temp_local_registry.mkdir()
        self.cli_options = (
            "--registry-path",
            str(self.temp_local_registry),
            "publish",
        )

        shutil.copytree(ROOT_DIR / "packages", self.packages_dir)
        os.chdir(self.service_dir)

    def test_local(
        self,
    ) -> None:
        """Test publish service locally."""
        result = self.run_cli(("--local",))

        assert result.exit_code == 0, result.output
        assert (
            'Service "counter" successfully published on the local packages directory.'
            in result.output
        )

    def test_ipfs(
        self,
    ) -> None:
        """Test publish service on IPFS registry."""

        dummy_hash_v0 = "QmeU1Cz796TBihCT426pA3HAYC7LhaawsXgGmy1hpyZXj9"
        dummy_hash_v1 = to_v1("QmeU1Cz796TBihCT426pA3HAYC7LhaawsXgGmy1hpyZXj9")

        with mock.patch(
            "autonomy.cli.helpers.registry.get_default_remote_registry",
            new=lambda: REMOTE_IPFS,
        ), mock.patch.object(
            IPFSTool,
            "add",
            new=lambda *_: (
                None,
                dummy_hash_v0,
                None,
            ),
        ):
            result = self.run_cli(("--remote",))

        msg_check = f"""Service "counter" successfully published on the IPFS registry.\n\tPublicId: valory/counter:0.1.0\n\tPackage hash: {dummy_hash_v1}"""
        assert result.exit_code == 0, result.output
        assert msg_check in result.output, result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `from-token` command."""

import json
import os
from pathlib import Path
from unittest import mock

import pytest
from aea.cli.registry.settings import REMOTE_IPFS
from aea_test_autonomy.fixture_helpers import registries_scope_class  # noqa: F401

from autonomy.chain.exceptions import FailedToRetrieveComponentMetadata

from tests.conftest import ROOT_DIR, skip_docker_tests
from tests.test_autonomy.test_chain.base import BaseChainInteractionTest


MOCK_IPFS_RESPONSE = {
    "name": "valory/oracle_hardhat",
    "description": "Oracle service.",
    "code_uri": "ipfs://bafybeiansmhkoovd6jlnyurm2w4qzhpmi43gxlyenq33ioovy2rh4gziji",
    "image": "bafybeiansmhkoovd6jlnyurm2w4qzhpmi43gxlyenq33ioovy2rh4gziji",
    "attributes": [{"trait_type": "version", "value": "0.1.0"}],
}
REGISTERED_KEYS = [
    {
        "address": "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        "private_key": "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80",
    },
    {
        "address": "0x70997970C51812dc3A010C7d01b50e0d17dc79C8",
        "private_key": "0x59c6995e998f97a5a0044966f0945389dc9e86dae88c7a8412f4603b6b78690d",
    },
    {
        "address": "0x3C44CdDdB6a900fa2b585dd299e03d12FA4293BC",
        "private_key": "0x5de4111afa1a4b94908f83103eb1f1706367c2e68ca870fc3fb9a804cdab365a",
    },
    {
        "address": "0x90F79bf6EB2c4f870365E785982E1f101E93b906",
        "private_key": "0x7c852118294e51e653712a81e05800f419141751be58f605c371e15141b007a6",
    },
]

run_deployment_patch = mock.patch("autonomy.cli.helpers.deployment.run_deployment")
build_image_patch = mock.patch("autonomy.cli.helpers.deployment.build_image")
default_remote_registry_patch = mock.patch(
    "autonomy.cli.helpers.registry.get_default_remote_registry",
    new=lambda: REMOTE_IPFS,
)
default_ipfs_node_patch = mock.patch(
    "autonomy.cli.helpers.registry.get_ipfs_node_multiaddr",
    new=lambda: "/dns/registry.autonolas.tech/tcp/443/https",
)
ipfs_resolve_patch = mock.patch(
    "autonomy.cli.helpers.deployment.resolve_component_id",
    return_value=MOCK_IPFS_RESPONSE,
)


@pytest.mark.integration
@skip_docker_tests
class TestFromToken(BaseChainInteractionTest):
    """Test `from-token` command."""

    cli_options = ("deploy", "from-token")
    token = 1
    chain = "staging"
    keys_file: Path

    def setup(self) -> None:
        """Setup test method."""
        super().setup()

        os.chdir(self.t)
        self.keys_file = self.t / "keys.json"
        self.keys_file.write_text(json.dumps(REGISTERED_KEYS))

    def test_from_token(
        self,
    ) -> None:
        """Run test."""

        service_dir = self.t / "service"
        service_dir.mkdir()

        service_file = service_dir / "service.yaml"
        service_file.write_text(
            (
                ROOT_DIR
                / "tests"
                / "data"
                / "dummy_service_config_files"
                / "service_0.yaml"
            ).read_text()
        )

        with mock.patch(
            "autonomy.cli.helpers.deployment.fetch_service_ipfs",
            return_value=service_dir,
        ), (
            run_deployment_patch
        ), (
            build_image_patch
        ), default_remote_registry_patch, default_ipfs_node_patch, ipfs_resolve_patch:
            result = self.run_cli(
                (
                    str(self.token),
                    str(self.keys_file),
                )
            )

            assert result.exit_code == 0, result.stdout
            assert "Service name: valory/oracle_hardhat" in result.stdout
            assert "Building required images" in result.stdout
            assert "Service build successful" in result.stdout

    def test_from_token_kubernetes(
        self,
    ) -> None:
        """Run test."""

        service_dir = self.t / "service"
        service_dir.mkdir()

        service_file = service_dir / "service.yaml"
        service_file.write_text(
            (
                ROOT_DIR
                / "tests"
                / "data"
                / "dummy_service_config_files"
                / "service_0.yaml"
            ).read_text()
        )

        with mock.patch(
            "autonomy.cli.helpers.deployment.fetch_service_ipfs",
            return_value=service_dir,
        ), run_deployment_patch as rdp, (
            build_image_patch
        ), default_remote_registry_patch, default_ipfs_node_patch, ipfs_resolve_patch:
            result = self.run_cli(
                (str(self.token), str(self.keys_file), "--kubernetes")
            )

            assert result.exit_code == 0, result.stdout
            assert "Service name: valory/oracle_hardhat" in result.stdout
            assert "Building required images" in result.stdout
            assert "Service build successful" in result.stdout
            assert "Type:                 kubernetes" in result.stdout
            assert "Running deployment" not in result.output

            rdp.assert_not_called()

        assert (self.t / "service" / "abci_build" / "build.yaml").exists()

    def test_fail_on_chain_resolve_connection_error(self) -> None:
        """Run test."""

        with (
            run_deployment_patch
        ), (
            build_image_patch
        ), (
            default_remote_registry_patch
        ), default_ipfs_node_patch, ipfs_resolve_patch, mock.patch(
            "autonomy.cli.helpers.deployment.resolve_component_id",
            side_effect=FailedToRetrieveComponentMetadata(
                "Error connecting RPC endpoint"
            ),
        ):
            result = self.run_cli(
                (
                    str(self.token),
                    str(self.keys_file),
                )
            )

            assert result.exit_code == 1, result.stdout
            assert "Error connecting RPC endpoint" in result.stderr, result.output

    def test_fail_on_chain_resolve_bad_contract_call(self) -> None:
        """Run test."""

        with (
            run_deployment_patch
        ), (
            build_image_patch
        ), (
            default_remote_registry_patch
        ), default_ipfs_node_patch, ipfs_resolve_patch, mock.patch(
            "autonomy.cli.helpers.deployment.resolve_component_id",
            side_effect=Exception,
        ):
            result = self.run_cli(
                (
                    str(self.token),
                    str(self.keys_file),
                )
            )

            assert result.exit_code == 1, result.stdout
            assert "Cannot find the service registry deployment;" in result.stderr



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for ipfs protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.valory.protocols.ipfs.message import IpfsMessage


class TestMessageIpfs(BaseProtocolMessagesTestCase):
    """Test for the 'ipfs' protocol message."""

    MESSAGE_CLASS = IpfsMessage

    def build_messages(self) -> List[IpfsMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            IpfsMessage(
                performative=IpfsMessage.Performative.STORE_FILES,
                files={"some str": "some str"},
                timeout=1.0,
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.IPFS_HASH,
                ipfs_hash="some str",
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.GET_FILES,
                ipfs_hash="some str",
                timeout=1.0,
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.FILES,
                files={"some str": "some str"},
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.ERROR,
                reason="some str",
            ),
        ]

    def build_inconsistent(self) -> List[IpfsMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            IpfsMessage(
                performative=IpfsMessage.Performative.STORE_FILES,
                # skip content: files
                timeout=1.0,
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.IPFS_HASH,
                # skip content: ipfs_hash
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.GET_FILES,
                # skip content: ipfs_hash
                timeout=1.0,
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.FILES,
                # skip content: files
            ),
            IpfsMessage(
                performative=IpfsMessage.Performative.ERROR,
                # skip content: reason
            ),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for abstract_round_abci/test_tools/base.py"""

from enum import Enum
from typing import Any, Dict, cast

import pytest
from aea.mail.base import Envelope

from packages.valory.connections.ledger.connection import (
    PUBLIC_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviours import BaseBehaviour
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    DummyContext,
    FSMBehaviourBaseCase,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci import PUBLIC_ID
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.behaviours import (
    DummyRoundBehaviour,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.models import (
    SharedState,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci.rounds import (
    Event,
    SynchronizedData,
)
from packages.valory.skills.abstract_round_abci.tests.test_tools.base import (
    FSMBehaviourTestToolSetup,
)


class TestFSMBehaviourBaseCaseSetup(FSMBehaviourTestToolSetup):
    """test TestFSMBehaviourBaseCaseSetup setup"""

    test_cls = FSMBehaviourBaseCase

    @pytest.mark.parametrize("kwargs", [{}])
    def test_setup_fails_without_path(self, kwargs: Dict[str, Dict[str, Any]]) -> None:
        """Test setup"""
        with pytest.raises(ValueError):
            self.test_cls.setup_class(**kwargs)

    @pytest.mark.parametrize("kwargs", [{}, {"param_overrides": {"new_p": None}}])
    def test_setup(self, kwargs: Dict[str, Dict[str, Any]]) -> None:
        """Test setup"""

        self.set_path_to_skill()
        test_instance = self.setup_test_cls(**kwargs)
        assert test_instance
        assert hasattr(test_instance.behaviour.context.params, "new_p") == bool(kwargs)

    @pytest.mark.parametrize("behaviour", DummyRoundBehaviour.behaviours)
    def test_fast_forward_to_behaviour(self, behaviour: BaseBehaviour) -> None:
        """Test fast_forward_to_behaviour"""
        self.set_path_to_skill()
        test_instance = self.setup_test_cls()

        skill = test_instance._skill  # pylint: disable=protected-access
        round_behaviour = skill.skill_context.behaviours.main
        behaviour_id = behaviour.behaviour_id
        synchronized_data = SynchronizedData(
            AbciAppDB(setup_data=dict(participants=[tuple("abcd")]))
        )

        test_instance.fast_forward_to_behaviour(
            behaviour=round_behaviour,
            behaviour_id=behaviour_id,
            synchronized_data=synchronized_data,
        )

        current_behaviour = test_instance.behaviour.current_behaviour
        assert current_behaviour is not None
        assert isinstance(
            current_behaviour.synchronized_data,
            SynchronizedData,
        )
        assert current_behaviour.behaviour_id == behaviour.behaviour_id
        assert (  # pylint: disable=protected-access
            test_instance.skill.skill_context.state.round_sequence.abci_app._current_round_cls
            == current_behaviour.matching_round
            == behaviour.matching_round
        )

    @pytest.mark.parametrize("event", Event)
    @pytest.mark.parametrize("set_none", [False, True])
    def test_end_round(self, event: Enum, set_none: bool) -> None:
        """Test end_round"""

        self.set_path_to_skill()
        test_instance = self.setup_test_cls()
        current_behaviour = cast(
            BaseBehaviour, test_instance.behaviour.current_behaviour
        )
        abci_app = current_behaviour.context.state.round_sequence.abci_app
        if set_none:
            test_instance.behaviour.current_behaviour = None
        assert abci_app.current_round_height == 0
        test_instance.end_round(event)
        assert abci_app.current_round_height == 1 - int(set_none)

    def test_mock_ledger_api_request(self) -> None:
        """Test mock_ledger_api_request"""

        self.set_path_to_skill()
        test_instance = self.setup_test_cls()

        request_kwargs = dict(performative=LedgerApiMessage.Performative.GET_BALANCE)
        response_kwargs = dict(performative=LedgerApiMessage.Performative.BALANCE)
        with pytest.raises(
            AssertionError,
            match="Invalid number of messages in outbox. Expected 1. Found 0.",
        ):
            test_instance.mock_ledger_api_request(request_kwargs, response_kwargs)

        message = LedgerApiMessage(**request_kwargs, dialogue_reference=("a", "b"))  # type: ignore
        envelope = Envelope(
            to=str(LEDGER_CONNECTION_PUBLIC_ID),
            sender=str(PUBLIC_ID),
            protocol_specification_id=LedgerApiMessage.protocol_specification_id,
            message=message,
        )
        multiplexer = test_instance._multiplexer  # pylint: disable=protected-access
        multiplexer.out_queue.put_nowait(envelope)
        test_instance.mock_ledger_api_request(request_kwargs, response_kwargs)

    def test_mock_contract_api_request(self) -> None:
        """Test mock_contract_api_request"""

        self.set_path_to_skill()
        test_instance = self.setup_test_cls()

        contract_id = "dummy_contract"
        request_kwargs = dict(performative=ContractApiMessage.Performative.GET_STATE)
        response_kwargs = dict(performative=ContractApiMessage.Performative.STATE)
        with pytest.raises(
            AssertionError,
            match="Invalid number of messages in outbox. Expected 1. Found 0.",
        ):
            test_instance.mock_contract_api_request(
                contract_id, request_kwargs, response_kwargs
            )

        message = ContractApiMessage(
            **request_kwargs,  # type: ignore
            dialogue_reference=("a", "b"),
            ledger_id="ethereum",
            contract_id=contract_id
        )
        envelope = Envelope(
            to=str(LEDGER_CONNECTION_PUBLIC_ID),
            sender=str(PUBLIC_ID),
            protocol_specification_id=ContractApiMessage.protocol_specification_id,
            message=message,
        )
        multiplexer = test_instance._multiplexer  # pylint: disable=protected-access
        multiplexer.out_queue.put_nowait(envelope)
        test_instance.mock_contract_api_request(
            contract_id, request_kwargs, response_kwargs
        )


def test_dummy_context_is_abstract_component() -> None:
    """Test dummy context is abstract component"""

    shared_state = SharedState(name="dummy_shared_state", skill_context=DummyContext())
    assert shared_state.context.is_abstract_component



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test for contract module."""

from pathlib import Path

from aea_test_autonomy.base_test_classes.contracts import BaseRegistriesContractsTest
from aea_test_autonomy.docker.base import skip_docker_tests
from aea_test_autonomy.docker.registries import REGISTRIES_MANAGER

from packages.valory.contracts.registries_manager.contract import (
    RegistriesManagerContract,
)


PACKAGE_DIR = Path(__file__).parent.parent

METADATA_HASH = "0xaaca27f6156089376fd85900d511c7570b6c9b0f6afc64576aac5aa0da8de92d"


@skip_docker_tests
class TestRegistriesManager(BaseRegistriesContractsTest):
    """Test registries manager."""

    contract: RegistriesManagerContract
    contract_address = REGISTRIES_MANAGER
    contract_directory = PACKAGE_DIR

    def test_get_create_transaction(self) -> None:
        """Test `get_create_transaction` method."""

        tx = self.contract.get_create_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            metadata_hash=METADATA_HASH,
            component_type=RegistriesManagerContract.UnitType.COMPONENT,
            owner=self.deployer_crypto.address,
            sender=self.deployer_crypto.address,
        )

        assert all(
            [
                key
                in [
                    "chainId",
                    "nonce",
                    "value",
                    "gas",
                    "maxFeePerGas",
                    "maxPriorityFeePerGas",
                    "to",
                    "data",
                ]
                for key in tx.keys()
            ]
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for abstract_round_abci/test_tools/common.py"""

from pathlib import Path
from typing import Any, Dict, Type, cast

from aea.helpers.base import cd
from aea.test_tools.utils import copy_class

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload, _MetaPayload
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.abstract_round_abci.tests.data.dummy_abci import (
    PATH_TO_SKILL,
)


class FSMBehaviourTestToolSetup:
    """BaseRandomnessBehaviourTestSetup"""

    test_cls: Type[FSMBehaviourBaseCase]
    __test_cls: Type[FSMBehaviourBaseCase]
    __old_value: Dict[str, Type[BaseTxPayload]]

    @classmethod
    def setup_class(cls) -> None:
        """Setup class"""

        if not hasattr(cls, "test_cls"):
            raise AttributeError(f"{cls} must set `test_cls`")

        cls.__test_cls = cls.test_cls
        cls.__old_value = _MetaPayload.registry.copy()
        _MetaPayload.registry.clear()

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown class"""
        _MetaPayload.registry = cls.__old_value

    def setup(self) -> None:
        """Setup test"""
        test_cls = copy_class(self.__test_cls)
        self.test_cls = cast(Type[FSMBehaviourBaseCase], test_cls)

    def teardown(self) -> None:
        """Teardown test"""
        self.test_cls.teardown_class()

    def set_path_to_skill(self, path_to_skill: Path = PATH_TO_SKILL) -> None:
        """Set path_to_skill"""
        self.test_cls.path_to_skill = path_to_skill

    def setup_test_cls(self, **kwargs: Any) -> FSMBehaviourBaseCase:
        """Helper method to setup test to be tested"""

        # different test tools will require the setting of
        # different class attributes (such as path_to_skill).
        # One should write a test that sets these,
        # and subsequently invoke this method to test the setup.

        with cd(self.test_cls.path_to_skill):
            self.test_cls.setup_class(**kwargs)

        test_instance = self.test_cls()
        test_instance.setup()
        return test_instance



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""The module contains test_helpers for module tests."""

import json
import os
import random
import string
import tempfile
from pathlib import Path
from unittest import mock

from aea.configurations.constants import DEFAULT_ENV_DOTFILE

from autonomy.cli.helpers.env import load_env_file


class TestLoadEnvFile:
    """Test `load_env_file` file."""

    _t: tempfile.TemporaryDirectory
    t: Path
    cwd: Path

    env_var = "ENV_VAR"

    @staticmethod
    def generate_random_value(n: int = 6) -> str:
        """Generate random string."""
        return "".join([random.choice(string.ascii_letters) for i in range(n)])  # nosec

    def setup(self) -> None:
        """Setup test"""

        self._t = tempfile.TemporaryDirectory()
        self.t = Path(self._t.name)
        self.cwd = Path.cwd()

        os.chdir(self.t)

    def test_load_dot_env(self) -> None:
        """Test `.env` file."""
        env_var_value = self.generate_random_value()
        dotenv_file = self.t / DEFAULT_ENV_DOTFILE
        dotenv_file.write_text(f"{self.env_var}={env_var_value}")
        with mock.patch.dict(os.environ):
            load_env_file(file=dotenv_file)
            assert os.environ[self.env_var] == env_var_value

    def test_load_json(self) -> None:
        """Test `.json` file."""

        env_var_value = self.generate_random_value()
        json_file = self.t / "env.json"
        json_file.write_text(json.dumps({self.env_var: env_var_value}))
        with mock.patch.dict(os.environ):
            load_env_file(file=json_file)
            assert os.environ[self.env_var] == env_var_value

    def test_load_json_serialize(self) -> None:
        """Test `.json` file with serialize values."""

        env_var_value = [self.generate_random_value(), self.generate_random_value()]
        json_file = self.t / "env.json"
        json_file.write_text(json.dumps({self.env_var: env_var_value}))

        with mock.patch.dict(os.environ):
            load_env_file(file=json_file, serialize_json=True)
            assert json.loads(os.environ[self.env_var]) == env_var_value

    def teardown(self) -> None:
        """Test teardown"""
        os.chdir(self.cwd)
        self._t.cleanup()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains a wrapper around Multicall2."""
import logging
from typing import Any, Callable, Dict, List, Tuple

from aea.common import JSONLike
from aea.configurations.base import PublicId
from aea.contracts.base import Contract
from aea.crypto.base import LedgerApi


PUBLIC_ID = PublicId.from_str("valory/multicall2:0.1.0")

_logger = logging.getLogger(
    f"aea.packages.{PUBLIC_ID.author}.contracts.{PUBLIC_ID.name}.contract"
)


class Multicall2Contract(Contract):
    """A wrapper for the MakerDAO Multicall2."""

    contract_id = PUBLIC_ID

    @classmethod
    def get_raw_transaction(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_RAW_TRANSACTION' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def get_raw_message(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> bytes:
        """
        Handler method for the 'GET_RAW_MESSAGE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def get_state(
        cls, ledger_api: LedgerApi, contract_address: str, **kwargs: Any
    ) -> JSONLike:
        """
        Handler method for the 'GET_STATE' requests.

        Implement this method in the sub class if you want
        to handle the contract requests manually.

        :param ledger_api: the ledger apis.
        :param contract_address: the contract address.
        :param kwargs: the keyword arguments.
        :return: the tx  # noqa: DAR202
        """
        raise NotImplementedError  # pragma: nocover

    @classmethod
    def encode_function_call(
        cls,
        ledger_api: LedgerApi,
        contract_instance: Any,
        fn_name: str,
        args: List[Any],
    ) -> Tuple[Dict[str, Any], Callable]:
        """
        Encode a function call.

        To be used with Multicall2.aggregate().

        :param ledger_api: the ledger api.
        :param contract_instance: an insntace of the contract whose call is getting encoded.
        :param fn_name: the function name.
        :param args: the function arguments.
        :return: the target address, the data, and the output decoder function.
        """
        data = contract_instance.encodeABI(fn_name=fn_name, args=args)
        for elem in contract_instance.abi:
            if elem.get("name", "") == fn_name:
                output_types = elem["outputs"]
                normalized_output_types = [t["type"] for t in output_types]
                break

        def decoder(return_data: bytes) -> Any:
            """Decode the return data."""
            return ledger_api.api.codec.decode(normalized_output_types, return_data)

        decoder_fn = decoder
        target_address = contract_instance.address
        call = {
            "target": target_address,
            "callData": data,
        }
        return call, decoder_fn

    @classmethod
    def aggregate_and_decode(
        cls,
        ledger_api: LedgerApi,
        contract_address: str,
        calls_and_decoders: List[Tuple[Dict[str, Any], Callable]],
    ) -> Tuple[int, List[Any]]:
        """
        Make aggregate call.

        :param ledger_api: the ledger apis.
        :param contract_address: the multicall address.
        :param calls_and_decoders: the calls and their corresponding output decoders.
        :return: block number and decoded outputs.
        """
        instance = cls.get_instance(ledger_api, contract_address)
        calls = [call[0] for call in calls_and_decoders]
        decoders = [call[1] for call in calls_and_decoders]
        res = instance.functions.aggregate(calls).call()
        block_number, call_responses = res[0], res[1]
        decoded_responses = [
            decoder(call_response)
            for decoder, call_response in zip(decoders, call_responses)
        ]
        block_number = ledger_api.api.eth.blockNumber
        return block_number, decoded_responses



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains all the loading operations of the behaviours."""

import json
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional

from packages.valory.skills.abstract_round_abci.io_.store import (
    CustomObjectType,
    NativelySupportedSingleObjectType,
    SupportedFiletype,
    SupportedObjectType,
    SupportedSingleObjectType,
)


CustomLoaderType = Optional[Callable[[str], CustomObjectType]]
SupportedLoaderType = Callable[[str], SupportedSingleObjectType]


class AbstractLoader(ABC):
    """An abstract `Loader` class."""

    @abstractmethod
    def load_single_object(
        self, serialized_object: str
    ) -> NativelySupportedSingleObjectType:
        """Load a single object."""

    def load(self, serialized_objects: Dict[str, str]) -> SupportedObjectType:
        """
        Load one or more serialized objects.

        :param serialized_objects: A mapping of filenames to serialized object they contained.
        :return: the loaded file(s).
        """
        if len(serialized_objects) == 0:
            # no objects are present, raise an error
            raise ValueError('"serialized_objects" does not contain any objects')

        objects = {}
        for filename, body in serialized_objects.items():
            objects[filename] = self.load_single_object(body)

        if len(objects) > 1:
            # multiple object are present
            # we return them as mapping of
            # names and their value
            return objects

        # one object is present, we simply return it as an object, i.e. without its name
        _name, deserialized_body = objects.popitem()
        return deserialized_body


class JSONLoader(AbstractLoader):
    """A JSON file loader."""

    def load_single_object(
        self, serialized_object: str
    ) -> NativelySupportedSingleObjectType:
        """Read a json file.

        :param serialized_object: the file serialized into a JSON string.
        :return: the deserialized json file's content.
        """
        try:
            deserialized_file = json.loads(serialized_object)
            return deserialized_file
        except json.JSONDecodeError as e:  # pragma: no cover
            raise IOError(
                f"File '{serialized_object}' has an invalid JSON encoding!"
            ) from e
        except ValueError as e:  # pragma: no cover
            raise IOError(
                f"There is an encoding error in the '{serialized_object}' file!"
            ) from e


class Loader(AbstractLoader):
    """Class which loads objects."""

    def __init__(self, filetype: Optional[Any], custom_loader: CustomLoaderType):
        """Initialize a `Loader`."""
        self._filetype = filetype
        self._custom_loader = custom_loader
        self.__filetype_to_loader: Dict[SupportedFiletype, SupportedLoaderType] = {
            SupportedFiletype.JSON: JSONLoader().load_single_object,
        }

    def load_single_object(self, serialized_object: str) -> SupportedSingleObjectType:
        """Load a single file."""
        loader = self._get_single_loader_from_filetype()
        return loader(serialized_object)

    def _get_single_loader_from_filetype(self) -> SupportedLoaderType:
        """Get an object loader from a given filetype or keep a custom loader."""
        if self._filetype is not None:
            return self.__filetype_to_loader[self._filetype]

        if self._custom_loader is not None:  # pragma: no cover
            return self._custom_loader

        raise ValueError(  # pragma: no cover
            "Please provide either a supported filetype or a custom loader function."
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the termination behaviour classes."""
import sys
from typing import Callable, Dict, Generator, List, Optional, Set, Type, cast

from aea.protocols.base import Message
from hexbytes import HexBytes

from packages.valory.contracts.gnosis_safe.contract import (
    GnosisSafeContract,
    SafeOperation,
)
from packages.valory.contracts.multisend.contract import (
    MultiSendContract,
    MultiSendOperation,
)
from packages.valory.contracts.service_registry.contract import ServiceRegistryContract
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    AsyncBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.behaviours import AbstractRoundBehaviour
from packages.valory.skills.termination_abci.models import TerminationParams
from packages.valory.skills.termination_abci.payloads import BackgroundPayload
from packages.valory.skills.termination_abci.rounds import (
    BackgroundRound,
    SynchronizedData,
    TerminationAbciApp,
    TerminationRound,
)
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    TransactionSettlementRoundBehaviour,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    hash_payload_to_hex,
)


# setting the safe gas to 0 means that all available gas will be used
# which is what we want in most cases
# more info here: https://safe-docs.dev.gnosisdev.com/safe/docs/contracts_tx_execution/
_SAFE_GAS = 0

# hardcoded to 0 because we don't need to send any ETH when handing over multisig ownership
_ETHER_VALUE = 0

# payload to represent a non-existing event
_NO_EVENT_FOUND: Dict = {}


class BackgroundBehaviour(BaseBehaviour):
    """A behaviour responsible for picking up the termination signal, it runs concurrently with other behaviours."""

    matching_round = BackgroundRound
    _service_owner_address: Optional[str] = None

    def async_act(self) -> Generator:
        """
        Performs the termination logic.

        This method responsible for checking whether the on-chain termination signal is present.
        When an agent picks up the termination signal, it prepares a multisend transaction,
        on which the majority of agents in the service have to reach consensus on.

        :return: None
        :yield: None
        """
        if not self._is_majority_possible() or self._is_termination_majority():
            # if the service has not enough participants to reach
            # consensus or if termination majority has already
            # been reached, ie the multisend transaction has been
            # prepared by enough participants,
            # there is no need to run the rest of the act
            return

        signal_present = yield from self.check_for_signal()
        if signal_present is None:
            # if the response is None, something went wrong
            self.context.logger.error("Failed checking the termination signal.")
            return
        if not signal_present:
            # the signal is not present, so we sleep and try again
            yield from self.sleep(self.params.termination_sleep)
            return

        self.context.logger.info(
            "Termination signal was picked up, preparing termination transaction."
        )

        background_data = yield from self.get_multisend_payload()
        if background_data is None:
            self.context.logger.error(
                "Couldn't prepare multisend transaction for termination."
            )
            return

        self.context.logger.info("Successfully prepared termination multisend tx.")
        termination_payload = BackgroundPayload(
            self.context.agent_address, background_data=background_data
        )
        yield from self.send_a2a_transaction(termination_payload)
        yield from self.wait_for_termination_majority()

    @property
    def synchronized_data(self) -> SynchronizedData:
        """
        Return the synchronized data.

        Note: we instantiate here, rather than cast, as this runs
        concurrently and so the instantiation needs to happen somewhere.
        """
        return SynchronizedData(db=super().synchronized_data.db)

    @property
    def params(self) -> TerminationParams:
        """Return the params."""
        return cast(TerminationParams, super().params)

    def check_for_signal(self) -> Generator[None, None, Optional[bool]]:
        """
        This method checks for the termination signal.

        We check for the signal by looking at the "SafeReceived" events on the safe contract.
        https://github.com/safe-global/safe-contracts/blob/v1.3.0/contracts/common/EtherPaymentFallback.sol#L10
        In order for an SafeReceived event to qualify as a termination signal it has to fulfill the following:
            1. It MUST be sent by the service owner (self._service_owner).
            2. It MUST be 0 value tx.

        :returns: True if the termination signal is found, false otherwise
        """
        if self._service_owner_address is None:
            self._service_owner_address = yield from self._get_service_owner()

        termination_signal = yield from self._get_latest_termination_signal()
        if termination_signal is None:
            # something went wrong, we stop executing the rest of the logic
            return None
        if termination_signal == _NO_EVENT_FOUND:
            # no termination signal has ever been sent to safe
            return False

        service_owner_removal = yield from self._get_latest_removed_owner_event()
        if service_owner_removal is None:
            # something went wrong, we stop executing the rest of the logic
            return None
        if service_owner_removal == _NO_EVENT_FOUND:
            # the service owner has never been removed from the safe
            # this means that the observed `termination_signal` is the
            # first termination signal ever sent to this safe contract
            # as such it is valid signal for terminating the service
            return True

        # if both the `termination_signal` and the `service_owner` removal events are present,
        # we need to make sure this is not a previous termination that is already taken care of
        # if a termination has been previously made, we assume that:
        #   1. the service owner has been previously set as the safe owner
        #   2. since the service is running again, the ownership of the safe has been handed back to the agent instances
        # for 2. to happen, the service owner needs to be removed from being an owner of the safe. When this is done,
        # a `RemovedOwner` event is thrown, which is what `_get_latest_removed_owner_event()` captures.

        termination_signal_occurrence = int(termination_signal["block_number"])
        service_owner_removal_occurrence = int(service_owner_removal["block_number"])

        # if the termination signal has occurred after the owner has been removed the service should terminate,
        # otherwise it's a signal that has already been handled previously
        return termination_signal_occurrence > service_owner_removal_occurrence

    def _get_latest_removed_owner_event(self) -> Generator[None, None, Optional[Dict]]:
        """Returns the latest event in which the service owner was removed from the set of owners of the safe."""
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_removed_owner_events",
            contract_address=self.synchronized_data.safe_contract_address,
            removed_owner=self._service_owner_address,
        )
        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get the latest `RemovedOwner` event. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        removed_owner_events = cast(List[Dict], response.state.body.get("data"))
        if len(removed_owner_events) == 0:
            return _NO_EVENT_FOUND

        latest_removed_owner_event = removed_owner_events[0]
        for removed_owner_event in removed_owner_events[1:]:
            if int(removed_owner_event["block_number"]) > int(
                latest_removed_owner_event["block_number"]
            ):
                latest_removed_owner_event = removed_owner_event

        return latest_removed_owner_event

    def _get_latest_termination_signal(self) -> Generator[None, None, Optional[Dict]]:
        """Get the latest termination signal sent by the service owner."""
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_zero_transfer_events",
            contract_address=self.synchronized_data.safe_contract_address,
            sender_address=self._service_owner_address,
        )
        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get the latest Zero Transfer (`SafeReceived`) event. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        zero_transfer_events = cast(List[Dict], response.state.body.get("data"))
        if len(zero_transfer_events) == 0:
            return _NO_EVENT_FOUND

        latest_zero_transfer_event = zero_transfer_events[0]
        for zero_transfer_event in zero_transfer_events[1:]:
            if int(zero_transfer_event["block_number"]) > int(
                latest_zero_transfer_event["block_number"]
            ):
                latest_zero_transfer_event = zero_transfer_event

        return latest_zero_transfer_event

    def _get_service_owner(self) -> Generator[None, None, Optional[str]]:
        """Method that returns the service owner."""
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_id=str(ServiceRegistryContract.contract_id),
            contract_callable="get_service_owner",
            contract_address=self.params.service_registry_address,
            service_id=self.params.on_chain_service_id,
        )

        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get the service owner for service with id={self.params.on_chain_service_id}. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        service_owner = cast(
            Optional[str], response.state.body.get("service_owner", None)
        )
        return service_owner

    def get_multisend_payload(self) -> Generator[None, None, Optional[str]]:
        """Prepares and returns the multisend to hand over safe ownership to the service_owner."""
        multisend_data_str = yield from self._get_multisend_tx()
        if multisend_data_str is None:
            return None
        multisend_data = bytes.fromhex(multisend_data_str)
        tx_hash = yield from self._get_safe_tx_hash(multisend_data)
        if tx_hash is None:
            return None

        payload_data = hash_payload_to_hex(
            safe_tx_hash=tx_hash,
            ether_value=_ETHER_VALUE,
            safe_tx_gas=_SAFE_GAS,
            operation=SafeOperation.DELEGATE_CALL.value,
            to_address=self.params.multisend_address,
            data=multisend_data,
        )
        return payload_data

    def _get_safe_owners(self) -> Generator[None, None, Optional[List[str]]]:
        """Retrieves safe owners."""
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_owners",
            contract_address=self.synchronized_data.safe_contract_address,
        )

        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get the safe owners for safe deployed at {self.synchronized_data.safe_contract_address}. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        owners = cast(Optional[List[str]], response.state.body.get("owners", None))
        return owners

    def _get_remove_owner_tx(
        self, owner: str, threshold: int
    ) -> Generator[None, None, Optional[str]]:
        """
        Gets a remove owner tx.

        :param owner: the owner to be removed.
        :param threshold: the new safe threshold to be set.
        :return: the tx data
        """
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_remove_owner_data",
            contract_address=self.synchronized_data.safe_contract_address,
            owner=owner,
            threshold=threshold,
        )

        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get a remove owner tx. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        # strip "0x" from the response
        tx_data = cast(str, response.state.body.get("data", None))[2:]
        return tx_data

    def _get_swap_owner_tx(
        self, old_owner: str, new_owner: str
    ) -> Generator[None, None, Optional[str]]:
        """
        Gets a swap owner tx.

        :param old_owner: the owner to be removed.
        :param new_owner: the new safe threshold to be set.
        :return: the tx data
        """
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_swap_owner_data",
            contract_address=self.synchronized_data.safe_contract_address,
            old_owner=old_owner,
            new_owner=new_owner,
        )

        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get a swap owner tx. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        # strip "0x" from the response
        tx_data = cast(str, response.state.body.get("data", None))[2:]
        return tx_data

    def _get_safe_tx_hash(self, data: bytes) -> Generator[None, None, Optional[str]]:
        """
        Prepares and returns the safe tx hash.

        :param data: the safe tx data.
        :return: the tx hash
        """
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_raw_safe_transaction_hash",
            to_address=self.params.multisend_address,
            value=_ETHER_VALUE,
            data=data,
            safe_tx_gas=_SAFE_GAS,
            operation=SafeOperation.DELEGATE_CALL.value,
        )

        if response.performative != ContractApiMessage.Performative.STATE:
            self.context.logger.error(
                f"Couldn't get safe hash. "
                f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        # strip "0x" from the response hash
        tx_hash = cast(str, response.state.body["tx_hash"])[2:]
        return tx_hash

    def _get_multisend_tx(self) -> Generator[None, None, Optional[str]]:
        """This method compiles a multisend transaction to give ownership of the safe contract to the service owner."""
        transactions: List[Dict] = []
        threshold = 1
        safe_owners = yield from self._get_safe_owners()
        if safe_owners is None:
            return None
        owner_to_be_swapped = safe_owners[0]
        # we remove all but one safe owner
        # reverse the list to avoid errors when removing owners
        # this is because the owners are stored as a linked list
        # in the safe, hence the order is important
        safe_owners = list(reversed(safe_owners[1:]))
        for owner in safe_owners:
            # we generate a tx to remove the current owner
            remove_tx_str = yield from self._get_remove_owner_tx(owner, threshold)
            if remove_tx_str is None:
                return None
            remove_tx = bytes.fromhex(remove_tx_str)
            # we append it to the list of the multisend txs
            transactions.append(
                {
                    "operation": MultiSendOperation.CALL,
                    "to": self.synchronized_data.safe_contract_address,
                    "value": _ETHER_VALUE,
                    "data": HexBytes(remove_tx),
                }
            )

        if owner_to_be_swapped != cast(str, self._service_owner_address):
            # if the service owner is not the owner to be swapped
            # we swap, otherwise it's not necessary
            swap_tx = yield from self._get_swap_owner_tx(
                owner_to_be_swapped,
                cast(str, self._service_owner_address),
            )
            if swap_tx is None:
                return None

            transactions.append(
                {
                    "operation": MultiSendOperation.CALL,
                    "to": self.synchronized_data.safe_contract_address,
                    "value": _ETHER_VALUE,
                    "data": HexBytes(swap_tx),
                }
            )

        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.params.multisend_address,
            contract_id=str(MultiSendContract.contract_id),
            contract_callable="get_tx_data",
            multi_send_txs=transactions,
        )
        if response.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
            self.context.logger.error(
                f"Couldn't compile the multisend tx. "
                f"Expected response performative {ContractApiMessage.Performative.RAW_TRANSACTION.value}, "  # type: ignore
                f"received {response.performative.value}."
            )
            return None

        # strip "0x" from the response
        multisend_data = cast(str, response.raw_transaction.body["data"])[2:]
        return multisend_data

    def wait_for_termination_majority(self) -> Generator:
        """
        Wait until we reach majority on the termination transaction.

        :yield: None
        """
        yield from self.wait_for_condition(self._is_termination_majority)

    def _is_termination_majority(self) -> bool:
        """Rely on the round to decide when majority is reached."""
        return self.synchronized_data.termination_majority_reached

    def _is_majority_possible(self) -> bool:
        """Checks whether the service has enough participants to reach consensus."""
        return (
            self.synchronized_data.nb_participants
            >= self.synchronized_data.consensus_threshold
        )

    def get_callback_request(self) -> Callable[[Message, "BaseBehaviour"], None]:
        """Wrapper for callback_request(), overridden to avoid mix-ups with normal (non-background) behaviours."""

        def callback_request(
            message: Message, _current_behaviour: BaseBehaviour
        ) -> None:
            """
            This method gets called when a response for a prior request is received.

            Overridden to remove the check that checks whether the behaviour that made the request is still active.
            The received message gets passed to the behaviour that invoked it, in this case it's always the
            background behaviour.

            :param message: the response.
            :param _current_behaviour: not used, left in to satisfy the interface.
            :return: none
            """
            if self.is_stopped:
                self.context.logger.info(
                    "dropping message as behaviour has stopped: %s", message
                )
                return

            if self.state != AsyncBehaviour.AsyncState.WAITING_MESSAGE:
                self.context.logger.warning(
                    f"could not send message {message} to {self.behaviour_id}"
                )
                return

            self.try_send(message)

        return callback_request

    def is_round_ended(self, round_id: str) -> Callable[[], bool]:
        """
        Get a callable to check whether the current round has ended.

        We consider the background round ended when we have majority on the termination transaction.
        Overriden to allow for the behaviour to send transactions at any time.

        :return: the termination majority callable
        """
        return self._is_termination_majority


class TerminationBehaviour(BaseBehaviour):
    """Behaviour responsible for terminating the agent."""

    matching_round = TerminationRound

    def async_act(self) -> Generator:
        """Logs termination and terminates."""
        self.context.logger.info("Terminating the agent.")
        sys.exit()
        yield


class TerminationAbciBehaviours(AbstractRoundBehaviour):
    """This class defines the behaviours required in terminating a service."""

    initial_behaviour_cls = TransactionSettlementRoundBehaviour.initial_behaviour_cls
    abci_app_cls = TerminationAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        BackgroundBehaviour,  # type: ignore
        TerminationBehaviour,  # type: ignore
        *TransactionSettlementRoundBehaviour.behaviours,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""Test the `rounds` test tool module of the skill."""

import re
from enum import Enum
from typing import Any, FrozenSet, Generator, List, Optional, Tuple, Type, cast
from unittest.mock import MagicMock

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from packages.valory.skills.abstract_round_abci.base import (
    AbciAppDB,
    BaseSynchronizedData,
)
from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
    BaseCollectDifferentUntilAllRoundTest,
    BaseCollectDifferentUntilThresholdRoundTest,
    BaseCollectSameUntilAllRoundTest,
    BaseCollectSameUntilThresholdRoundTest,
    BaseOnlyKeeperSendsRoundTest,
    BaseRoundTestClass,
    BaseVotingRoundTest,
    DummyCollectDifferentUntilAllRound,
    DummyCollectDifferentUntilThresholdRound,
    DummyCollectSameUntilAllRound,
    DummyCollectSameUntilThresholdRound,
    DummyEvent,
    DummyOnlyKeeperSendsRound,
    DummySynchronizedData,
    DummyTxPayload,
    DummyVotingRound,
    MAX_PARTICIPANTS,
    get_dummy_tx_payloads,
    get_participants,
)
from packages.valory.skills.abstract_round_abci.tests.conftest import profile_name
from packages.valory.skills.abstract_round_abci.tests.test_common import last_iteration


settings.load_profile(profile_name)

# this is how many times we need to iterate before reaching the last iteration for a base test.
BASE_TEST_GEN_ITERATIONS = 4


def test_get_participants() -> None:
    """Test `get_participants`."""
    participants = get_participants()
    assert isinstance(participants, frozenset)
    assert all(isinstance(p, str) for p in participants)
    assert len(participants) == MAX_PARTICIPANTS


class DummyTxPayloadMatcher:
    """A `DummyTxPayload` matcher for assertion comparisons."""

    expected: DummyTxPayload

    def __init__(self, expected: DummyTxPayload) -> None:
        """Initialize the matcher."""
        self.expected = expected

    def __repr__(self) -> str:
        """Needs to be implemented for better assertion messages."""
        return (
            "DummyTxPayload("
            f"id={repr(self.expected.id_)}, "
            f"round_count={repr(self.expected.round_count)}, "
            f"sender={repr(self.expected.sender)}, "
            f"value={repr(self.expected.value)}, "
            f"vote={repr(self.expected.vote)}"
            ")"
        )

    def __eq__(self, other: Any) -> bool:
        """The method that will be used for the assertion comparisons."""
        return (
            self.expected.round_count == other.round_count
            and self.expected.sender == other.sender
            and self.expected.value == other.value
            and self.expected.vote == other.vote
        )


@given(
    st.frozensets(st.text(max_size=200), max_size=100),
    st.text(max_size=500),
    st.one_of(st.none(), st.booleans()),
    st.booleans(),
)
def test_get_dummy_tx_payloads(
    participants: FrozenSet[str],
    value: str,
    vote: Optional[bool],
    is_value_none: bool,
) -> None:
    """Test `get_dummy_tx_payloads`."""
    expected = [
        DummyTxPayloadMatcher(
            DummyTxPayload(
                sender=agent,
                value=(value or agent) if not is_value_none else value,
                vote=vote,
            )
        )
        for agent in sorted(participants)
    ]

    actual = get_dummy_tx_payloads(participants, value, vote, is_value_none)

    assert len(actual) == len(expected) == len(participants)
    assert actual == expected


class TestDummyTxPayload:  # pylint: disable=too-few-public-methods
    """Test class for `DummyTxPayload`"""

    @staticmethod
    @given(st.text(max_size=200), st.text(max_size=500), st.booleans())
    def test_properties(
        sender: str,
        value: str,
        vote: bool,
    ) -> None:
        """Test all the properties."""
        dummy_tx_payload = DummyTxPayload(sender, value, vote)
        assert dummy_tx_payload.value == value
        assert dummy_tx_payload.vote == vote
        assert dummy_tx_payload.data == {"value": value, "vote": vote}


class TestDummySynchronizedData:  # pylint: disable=too-few-public-methods
    """Test class for `DummySynchronizedData`."""

    @staticmethod
    @given(st.lists(st.text(max_size=200), max_size=100))
    def test_most_voted_keeper_address(
        most_voted_keeper_address_data: List[str],
    ) -> None:
        """Test `most_voted_keeper_address`."""
        most_voted_keeper_address_key = "most_voted_keeper_address"

        dummy_synchronized_data = DummySynchronizedData(
            db=AbciAppDB(
                setup_data={
                    most_voted_keeper_address_key: most_voted_keeper_address_data
                }
            )
        )

        if len(most_voted_keeper_address_data) == 0:
            with pytest.raises(
                ValueError,
                match=re.escape(
                    f"'{most_voted_keeper_address_key}' "
                    "field is not set for this period [0] and no default value was provided.",
                ),
            ):
                _ = dummy_synchronized_data.most_voted_keeper_address
            return

        assert (
            dummy_synchronized_data.most_voted_keeper_address
            == most_voted_keeper_address_data[-1]
        )


class TestBaseRoundTestClass:
    """Test `BaseRoundTestClass`."""

    @staticmethod
    def test_test_no_majority_event() -> None:
        """Test `_test_no_majority_event`."""
        base_round_test = BaseRoundTestClass()
        base_round_test._event_class = DummyEvent  # pylint: disable=protected-access

        base_round_test._test_no_majority_event(  # pylint: disable=protected-access
            MagicMock(
                end_block=lambda: (
                    MagicMock(),
                    DummyEvent.NO_MAJORITY,
                )
            )
        )

    @staticmethod
    @given(st.integers(min_value=0, max_value=100), st.integers(min_value=1))
    def test_complete_run(iter_count: int, shift: int) -> None:
        """Test `_complete_run`."""

        def dummy_gen() -> Generator[MagicMock, None, None]:
            """A dummy generator."""
            return (MagicMock() for _ in range(iter_count))

        # test with the same number as the generator's contents
        gen = dummy_gen()
        BaseRoundTestClass._complete_run(  # pylint: disable=protected-access
            gen, iter_count
        )

        # assert that the generator has been fully consumed
        with pytest.raises(StopIteration):
            next(gen)

        # test with a larger count than a generator's
        with pytest.raises(StopIteration):
            BaseRoundTestClass._complete_run(  # pylint: disable=protected-access
                dummy_gen(), iter_count + shift
            )


class BaseTestBase:
    """Base class for the Base tests."""

    gen: Generator
    base_round_test: BaseRoundTestClass
    base_round_test_cls: Type[BaseRoundTestClass]
    test_method_name = "_test_round"

    def setup(self) -> None:
        """Setup that is run before each test."""
        self.base_round_test = self.base_round_test_cls()
        self.base_round_test._synchronized_data_class = (  # pylint: disable=protected-access
            DummySynchronizedData
        )
        self.base_round_test.setup()
        self.base_round_test._event_class = (  # pylint: disable=protected-access
            DummyEvent
        )

    def create_test_gen(self, **kwargs: Any) -> None:
        """Create the base test generator."""
        test_method = getattr(self.base_round_test, self.test_method_name)
        self.gen = test_method(**kwargs)

    def exhaust_base_test_gen(self) -> None:
        """Exhaust the base test generator."""
        for _ in range(BASE_TEST_GEN_ITERATIONS):
            next(self.gen)
        last_iteration(self.gen)

    def run_test(self, **kwargs: Any) -> None:
        """Run a test for a base test."""
        self.create_test_gen(**kwargs)
        self.exhaust_base_test_gen()


class DummyCollectDifferentUntilAllRoundWithEndBlock(
    DummyCollectDifferentUntilAllRound
):
    """A `DummyCollectDifferentUntilAllRound` with `end_block` implemented."""

    def __init__(self, dummy_exit_event: DummyEvent, *args: Any, **kwargs: Any):
        """Initialize the dummy class."""
        super().__init__(*args, **kwargs)
        self.dummy_exit_event = dummy_exit_event

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """A dummy `end_block` implementation."""
        if self.collection_threshold_reached and self.dummy_exit_event is not None:
            return (
                cast(
                    DummySynchronizedData,
                    self.synchronized_data.update(
                        most_voted_keeper_address=list(self.collection.keys())
                    ),
                ),
                self.dummy_exit_event,
            )
        return None


class TestBaseCollectDifferentUntilAllRoundTest(BaseTestBase):
    """Test `BaseCollectDifferentUntilAllRoundTest`."""

    base_round_test: BaseCollectDifferentUntilAllRoundTest
    base_round_test_cls = BaseCollectDifferentUntilAllRoundTest

    @given(
        st.one_of(st.none(), st.sampled_from(DummyEvent)),
    )
    def test_test_round(self, exit_event: DummyEvent) -> None:
        """Test `_test_round`."""
        test_round = DummyCollectDifferentUntilAllRoundWithEndBlock(
            exit_event,
            self.base_round_test.synchronized_data,
        )
        round_payloads = [
            DummyTxPayload(f"agent_{i}", str(i)) for i in range(MAX_PARTICIPANTS)
        ]
        synchronized_data_attr_checks = [
            lambda _synchronized_data: _synchronized_data.most_voted_keeper_address
        ]

        self.run_test(
            test_round=test_round,
            round_payloads=round_payloads,
            synchronized_data_update_fn=lambda synchronized_data, _: synchronized_data.update(
                most_voted_keeper_address=[
                    f"agent_{i}" for i in range(MAX_PARTICIPANTS)
                ]
            ),
            synchronized_data_attr_checks=synchronized_data_attr_checks,
            exit_event=exit_event,
        )


class DummyCollectSameUntilAllRoundWithEndBlock(DummyCollectSameUntilAllRound):
    """A `DummyCollectSameUntilAllRound` with `end_block` implemented."""

    def __init__(self, dummy_exit_event: DummyEvent, *args: Any, **kwargs: Any):
        """Initialize the dummy class."""
        super().__init__(*args, **kwargs)
        self.dummy_exit_event = dummy_exit_event

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """A dummy `end_block` implementation."""
        if self.collection_threshold_reached:
            return (
                cast(
                    DummySynchronizedData,
                    self.synchronized_data.update(
                        most_voted_keeper_address=self.common_payload
                    ),
                ),
                self.dummy_exit_event,
            )
        return None


class TestBaseCollectSameUntilAllRoundTest(BaseTestBase):
    """Test `BaseCollectSameUntilAllRoundTest`."""

    base_round_test: BaseCollectSameUntilAllRoundTest
    base_round_test_cls: Type[
        BaseCollectSameUntilAllRoundTest
    ] = BaseCollectSameUntilAllRoundTest

    @given(
        st.sampled_from(DummyEvent),
        st.text(max_size=500),
        st.booleans(),
    )
    def test_test_round(
        self, exit_event: DummyEvent, common_value: str, finished: bool
    ) -> None:
        """Test `_test_round`."""
        test_round = DummyCollectSameUntilAllRoundWithEndBlock(
            exit_event,
            self.base_round_test.synchronized_data,
        )
        round_payloads = {
            f"test{i}": DummyTxPayload(f"agent_{i}", common_value)
            for i in range(MAX_PARTICIPANTS)
        }
        synchronized_data_attr_checks = [
            lambda _synchronized_data: _synchronized_data.most_voted_keeper_address
        ]

        self.run_test(
            test_round=test_round,
            round_payloads=round_payloads,
            synchronized_data_update_fn=lambda synchronized_data, _: synchronized_data.update(
                most_voted_keeper_address=common_value
            ),
            synchronized_data_attr_checks=synchronized_data_attr_checks,
            most_voted_payload=common_value,
            exit_event=exit_event,
            finished=finished,
        )


class DummyCollectSameUntilThresholdRoundWithEndBlock(
    DummyCollectSameUntilThresholdRound
):
    """A `DummyCollectSameUntilThresholdRound` with `end_block` overriden."""

    def __init__(self, dummy_exit_event: DummyEvent, *args: Any, **kwargs: Any):
        """Initialize the dummy class."""
        super().__init__(*args, **kwargs)
        self.dummy_exit_event = dummy_exit_event

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """A dummy `end_block` override."""
        if self.threshold_reached:
            return (
                cast(
                    DummySynchronizedData,
                    self.synchronized_data.update(
                        most_voted_keeper_address=self.most_voted_payload
                    ),
                ),
                self.dummy_exit_event,
            )
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, DummyEvent.NO_MAJORITY
        return None


class TestBaseCollectSameUntilThresholdRoundTest(BaseTestBase):
    """Test `BaseCollectSameUntilThresholdRoundTest`."""

    base_round_test: BaseCollectSameUntilThresholdRoundTest
    base_round_test_cls: Type[
        BaseCollectSameUntilThresholdRoundTest
    ] = BaseCollectSameUntilThresholdRoundTest

    @given(
        st.sampled_from(DummyEvent),
        st.text(max_size=500),
    )
    def test_test_round(self, exit_event: DummyEvent, most_voted_payload: str) -> None:
        """Test `_test_round`."""
        test_round = DummyCollectSameUntilThresholdRoundWithEndBlock(
            exit_event,
            self.base_round_test.synchronized_data,
        )
        round_payloads = {
            f"test{i}": DummyTxPayload(f"agent_{i}", most_voted_payload)
            for i in range(MAX_PARTICIPANTS)
        }
        synchronized_data_attr_checks = [
            lambda _synchronized_data: _synchronized_data.most_voted_keeper_address
        ]

        self.run_test(
            test_round=test_round,
            round_payloads=round_payloads,
            synchronized_data_update_fn=lambda synchronized_data, _: synchronized_data.update(
                most_voted_keeper_address=most_voted_payload
            ),
            synchronized_data_attr_checks=synchronized_data_attr_checks,
            most_voted_payload=most_voted_payload,
            exit_event=exit_event,
        )


class DummyOnlyKeeperSendsRoundTest(DummyOnlyKeeperSendsRound):
    """A `DummyOnlyKeeperSendsRound` with `end_block` implemented."""

    def __init__(self, dummy_exit_event: DummyEvent, *args: Any, **kwargs: Any):
        """Initialize the dummy class."""
        super().__init__(*args, **kwargs)
        self.dummy_exit_event = dummy_exit_event

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """A dummy `end_block` implementation."""
        if self.keeper_payload is not None and any(
            [val is not None for val in self.keeper_payload.values]
        ):
            return (
                cast(
                    DummySynchronizedData,
                    self.synchronized_data.update(
                        blacklisted_keepers=self.keeper_payload.values[0]
                    ),
                ),
                self.dummy_exit_event,
            )
        return None


class TestBaseOnlyKeeperSendsRoundTest(BaseTestBase):
    """Test `BaseOnlyKeeperSendsRoundTest`."""

    base_round_test: BaseOnlyKeeperSendsRoundTest
    base_round_test_cls: Type[
        BaseOnlyKeeperSendsRoundTest
    ] = BaseOnlyKeeperSendsRoundTest
    most_voted_keeper_address: str = "agent_0"

    def setup(self) -> None:
        """Setup that is run before each test."""
        super().setup()
        self.base_round_test.synchronized_data.update(
            most_voted_keeper_address=self.most_voted_keeper_address
        )

    @given(
        st.sampled_from(DummyEvent),
        st.text(),
    )
    def test_test_round(self, exit_event: DummyEvent, keeper_value: str) -> None:
        """Test `_test_round`."""
        test_round = DummyOnlyKeeperSendsRoundTest(
            exit_event,
            self.base_round_test.synchronized_data,
        )
        keeper_payload = DummyTxPayload(self.most_voted_keeper_address, keeper_value)
        synchronized_data_attr_checks = [
            lambda _synchronized_data: _synchronized_data.blacklisted_keepers
        ]

        self.run_test(
            test_round=test_round,
            keeper_payloads=keeper_payload,
            synchronized_data_update_fn=lambda synchronized_data, _: synchronized_data.update(
                blacklisted_keepers=keeper_value
            ),
            synchronized_data_attr_checks=synchronized_data_attr_checks,
            exit_event=exit_event,
        )


class DummyBaseVotingRoundTestWithEndBlock(DummyVotingRound):
    """A `DummyVotingRound` with `end_block` overriden."""

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """A dummy `end_block` override."""
        if self.positive_vote_threshold_reached:
            synchronized_data = cast(
                DummySynchronizedData,
                self.synchronized_data.update(
                    is_keeper_set=bool(self.collection),
                ),
            )
            return synchronized_data, DummyEvent.DONE
        if self.negative_vote_threshold_reached:
            return self.synchronized_data, DummyEvent.NEGATIVE
        if self.none_vote_threshold_reached:
            return self.synchronized_data, DummyEvent.NONE
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, DummyEvent.NO_MAJORITY
        return None


class TestBaseVotingRoundTest(BaseTestBase):
    """Test `BaseVotingRoundTest`."""

    base_round_test: BaseVotingRoundTest
    base_round_test_cls: Type[BaseVotingRoundTest] = BaseVotingRoundTest

    @given(
        st.one_of(st.none(), st.booleans()),
    )
    def test_test_round(self, is_keeper_set: Optional[bool]) -> None:
        """Test `_test_round`."""
        if is_keeper_set is None:
            exit_event = DummyEvent.NONE
            self.test_method_name = "_test_voting_round_none"
        elif is_keeper_set:
            exit_event = DummyEvent.DONE
            self.test_method_name = "_test_voting_round_positive"
        else:
            exit_event = DummyEvent.NEGATIVE
            self.test_method_name = "_test_voting_round_negative"

        test_round = DummyBaseVotingRoundTestWithEndBlock(
            self.base_round_test.synchronized_data,
        )
        round_payloads = {
            f"test{i}": DummyTxPayload(f"agent_{i}", value="", vote=is_keeper_set)
            for i in range(MAX_PARTICIPANTS)
        }
        synchronized_data_attr_checks = [
            lambda _synchronized_data: _synchronized_data.is_keeper_set
        ]

        self.run_test(
            test_round=test_round,
            round_payloads=round_payloads,
            synchronized_data_update_fn=lambda synchronized_data, _: synchronized_data.update(
                is_keeper_set=is_keeper_set
            ),
            synchronized_data_attr_checks=synchronized_data_attr_checks,
            exit_event=exit_event,
        )


class DummyCollectDifferentUntilThresholdRoundWithEndBlock(
    DummyCollectDifferentUntilThresholdRound
):
    """A `DummyCollectDifferentUntilThresholdRound` with `end_block` implemented."""

    def __init__(self, dummy_exit_event: DummyEvent, *args: Any, **kwargs: Any):
        """Initialize the dummy class."""
        super().__init__(*args, **kwargs)
        self.dummy_exit_event = dummy_exit_event

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """A dummy `end_block` implementation."""
        if self.collection_threshold_reached and self.dummy_exit_event is not None:
            return (
                cast(
                    DummySynchronizedData,
                    self.synchronized_data.update(
                        most_voted_keeper_address=list(self.collection.keys())
                    ),
                ),
                self.dummy_exit_event,
            )
        return None


class TestBaseCollectDifferentUntilThresholdRoundTest(BaseTestBase):
    """Test `BaseCollectDifferentUntilThresholdRoundTest`."""

    base_round_test: BaseCollectDifferentUntilThresholdRoundTest
    base_round_test_cls: Type[
        BaseCollectDifferentUntilThresholdRoundTest
    ] = BaseCollectDifferentUntilThresholdRoundTest

    @given(st.sampled_from(DummyEvent))
    def test_test_round(self, exit_event: DummyEvent) -> None:
        """Test `_test_round`."""
        test_round = DummyCollectDifferentUntilThresholdRoundWithEndBlock(
            exit_event,
            self.base_round_test.synchronized_data,
        )
        round_payloads = {
            f"test{i}": DummyTxPayload(f"agent_{i}", str(i))
            for i in range(MAX_PARTICIPANTS)
        }
        synchronized_data_attr_checks = [
            lambda _synchronized_data: _synchronized_data.most_voted_keeper_address
        ]

        self.run_test(
            test_round=test_round,
            round_payloads=round_payloads,
            synchronized_data_update_fn=lambda synchronized_data, _: synchronized_data.update(
                most_voted_keeper_address=[
                    f"agent_{i}" for i in range(MAX_PARTICIPANTS)
                ]
            ),
            synchronized_data_attr_checks=synchronized_data_attr_checks,
            exit_event=exit_event,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `service-registry-network` command."""

import multiprocessing
import os
import time
from typing import Tuple, cast

import docker
import pytest
import requests
from _pytest.capture import CaptureFixture  # type: ignore

from autonomy.constants import (
    SERVICE_REGISTRY_CONTRACT_CONTAINER_NAME as CONTAINER_NAME,
)

from tests.conftest import skip_docker_tests
from tests.test_autonomy.test_cli.base import BaseCliTest


@pytest.mark.integration
@skip_docker_tests
class TestRunServiceLocally(BaseCliTest):
    """Test `service-registry-network` command."""

    cli_options: Tuple[str, ...] = (
        "develop",
        "service-registry-network",
    )
    expected_network_address = "http://localhost:8545"
    running_process: multiprocessing.Process

    def setup(self) -> None:
        """Setup test."""
        super().setup()
        os.chdir(self.t)

    def test_run_service_locally(
        self,
    ) -> None:
        """Run test."""

        self.running_process = self._invoke_command()
        max_retries = 30
        timeout = 5
        expected = (
            "Deploying contracts with the account: 0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
            "Account balance: 10000000000000000000000",
            "Owner of created components and agents: 0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
            "ComponentRegistry deployed to: 0x5FbDB2315678afecb367f032d93F642f64180aa3",
            "AgentRegistry deployed to: 0xe7f1725E7734CE288F8367e1Bb143E90bb3F0512",
            "RegistriesManager deployed to: 0x9fE46736679d2D9a65F0992F2272dE9f3c7fa6e0",
            "ServiceRegistry deployed to: 0x998abeb3E57409262aE5b751f60747921B33613E",
            "ServiceManager deployed to: 0x70e0bA845a1A0F2DA3359C97E0285013525FFC49",
            "Gnosis Safe master copy deployed to: 0x4826533B4897376654Bb4d4AD88B7faFD0C98528",
            "Gnosis Safe proxy factory deployed to: 0x99bbA657f2BbC93c02D617f8bA121cB8Fc104Acf",
            "Gnosis Safe Multisig deployed to: 0x0E801D84Fa97b50751Dbf25036d067dCf18858bF",
            "Gnosis Safe Multisig with same address deployed to: 0x8f86403A4DE0BB5791fa46B8e795C547942fE4Cf",
            "Gnosis Safe Multisend deployed to: 0x9d4454B023096f34B160D6B654540c56A1F81688",
        )

        for _ in range(max_retries):
            try:
                res = requests.get(self.expected_network_address)
                assert res.status_code == 200, "bad response from the network"
                # we return in this case
                self._stop_cli_process()
                captured = cast(CaptureFixture, self.cli_runner.capfd).readouterr()
                missing = set(expected).difference(captured.out.split("\n"))
                assert not missing, missing
                return
            except requests.ConnectionError:
                # the container is not yet available
                time.sleep(timeout)

        self._stop_cli_process()
        raise AssertionError(f"failed to start network after {max_retries} tries")

    def _stop_cli_process(
        self,
    ) -> None:
        """Kill the process running the CLI."""
        if self.running_process is not None:
            self.running_process.terminate()
            client = docker.from_env()
            client.containers.get(CONTAINER_NAME).kill()
            client.containers.get(CONTAINER_NAME).remove()

    def _invoke_command(self) -> multiprocessing.Process:
        """Run the command on a different process, as it blocks."""
        process = multiprocessing.Process(target=self.run_cli)
        process.start()
        return process
